#
#	ASCORE.AWK
#
#	Creates a Csound score from an "ASCORE" script
#
#
#  -----------------------------------------------------------------------
#  Copyright 2004-2006  Alfred Steffens Jr.
#  -----------------------------------------------------------------------
#  -----------------------------------------------------------------------
#    Copying Permission:
#
#    Ascore is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation.
#
#    Ascore is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with Ascore (see file called "COPYING"); if not, write to the Free
#    Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
#    02111-1307  USA
#  ------------------------------------------------------------------------
#
#	Version descriptions
#
#	ver. 3.5	10/09/05	Alfred Steffens Jr.
#		modified because gawk version 3.1.5 was "fixed" by the developers
#		so that a dash inside a range [+-0-9], meaning + or - 0 through 9
#		should be "correctly" seen as a fatal error (the dash after the +
#		should be escaped).  Prior versions (specifically 3.1.3) were
#		evidently smart enough to see that 0-9 was the range, not +-0.  This
#		is what happens when we "fix" something that isn't broke (attention
#		C++ programmers)
#		changed ERROR statements to print to stderr instead of stdout
#		and changed exit status on errors to 1
#	ver. 3.4	09/27/05	Alfred Steffens Jr.
#		bug fix to setting gain on tempo relation symbol, &
#	ver. 3.3	07/17/05	Alfred Steffens Jr.
#		mod to allow comments at the end of a child (definition) line
#	ver. 3.2	07/16/05	Alfred Steffens Jr.
#		mod to allow comments at the end of a REVERB (definition) line
#	ver. 3.1	07/04/05	Alfred Steffens Jr.
#		new capability: added "alias" substitutions
#		new capability: added SKIPTRACK mechanism
#		new capability: added RANDOMIZE statement
#	ver. 3.0b	10/09/04	Alfred Steffens Jr.
#		child statement -- bug fix in "increment_name"
#	ver. 3.0	09/19/04	Alfred Steffens Jr.
#		child statement
#	ver. 2.8	09/15/04	Alfred Steffens Jr.	bug fix
#		fixed a bug in the start time algebra "2&+[0.2q]  or @+[3*] etc"
#		added a debug in Sc_EvalCol -- when column sizes don't match show
#		what's in the column to help debugging
#	ver. 2.7	07/05/04	Alfred Steffens Jr.	bug fix
#		fixed (again) "$variable" syntax to handle results of "eval"
#		statements and with multiple args on a line
#		Also, added ability to concatenate multiples "$multiple1" "$multiple2"...
#	ver. 2.6	07/04/04	Alfred Steffens Jr.	bug fix
#		make SKIPSCORE ignore the RVB_ON and RVN_OFF
#	ver. 2.5	06/20/04	Alfred Steffens Jr.	bug fix
#		to allow "$variable" syntax in the "set" statement with more
#		than one argument (a list)
#	ver. 2.4	05/15/04	Alfred Steffens Jr.	changed
#		"string_matrix_to_csound" to allow non-integer instrument
#		numbers (for chords)
#	ver. 2.3,  01/03/04,	Alfred Steffens Jr., added "seteval", 
#		"Eval_Variable"
#	ver. 2.2,  12/19/03,	Alfred Steffens Jr.
#	ver. 2.1,  12/16/03,	Alfred Steffens Jr.
#
#
#
#-------------------------------------------------------------------------
#
#	What is Ascore?
#	---------------
#
#	Ascore is a user-interface for building Csound score files.  It is a
#	command-line, non-GUI interface but tries to provide a visual
#	representation of the score sequence in the Ascii-Art sense.  Ascore
#	is an AWK script.  To compile an Ascore score file, one simply
#	types the following command:
#
#		cat my_score_file | awk -f ascore.awk
#
#	from which Ascore produces a standard-named file called "hold.sco".
#	AWK is an ancient, venerable utility that exists an all unix computers
#	and unix-like computers.  For windows people, AWK is available for
#	free from the CYGWIN project: just download awk.exe and cygwin1.dll
#	into your current directory.  I don't know for certain, but it is
#	most likely that a port of AWK exists for the Mac and OS/2 and other
#	operating systems that I don't know about.  Therefore, Ascore is
#	practically platform-independent.
#
#
#	A "score" in Ascore is a number of columns in which each column
#	is a parallel "sound track" with time increasing as the column is read
#	down the page.  In this way you can see the relative positions in
#	time of the various passages in the score.  Individual notes are
#	not allowed in the track section.  Instead any object scored must
#	be a note pattern, i.e., a macro of predefined notes.  For example,
#
#	7*		Lovely_Melody		%			%
#	*		|					%			Trumpet_Notes
#	3q		%					Strings		|
#	q		Lovely_Melody		|			%
#	5*		|					%			%
#
#	The above portion of an Ascore "score" section contains 3 "tracks",
#	being the 3 columns starting with Lovely_Melody.  The leftmost
#	column is the score's timing track which must consist only of note
#	duration symbols.  This structure gives a crude pictorial view of
#	the music as it occurs in time.  For instance, Lovely_Melody is
#	a note pattern (previous defined--see below) that starts first.
#	After 7 eighth notes have elapsed, Trumpet_Notes begins.  After
#	another eighth note has elapsed, Strings begins.  A "|" symbol means
#	that the note pattern above it is still playing.  A "%" symbol is
#	a rest symbol.  At any row in the score, the current time is given
#	by the sum of all duration symbols in the rows above it.  Thus,
#	when Lovely_Melody starts for the second time above 8 eighth notes
#	(7* = 7 eighth notes, * = 1 eighth note) and 3 quarter notes (3q =
#	3 quarter notes) have already elapsed.
#
#	It will be seen from the above example that Ascore is a scoring
#	utility built around the idea of note patterns and visual sound
#	tracks.
#
#
#	Time Symbols
#	-------------
#
#	One of the most difficult aspects of working with raw Csound scores
#	is keeping track of the time.  In Ascore you keep track of time
#	with relative time symbols.  These symbols are musical notes durations
#	such as quarter notes, half notes, etc.  You don't have to calculate
#	the time for any of these symbols.  They are calculated for you based
#	on the value of the global TEMPO statement that must be at the top
#	of your file.  In the timing track of the score section the current
#	time in the score is the sum of all the time duration symbols that
#	precede the current row.  In the above example there are five rows.
#	If you append a sixth row at the bottom of the score its note pattern
#	will begin at time (7*) + (*) + (3q) + (q) + (5*), which is 7 eighth
#	notes plus 1 eighth note plus 3 quarter notes plus 1 quarter note
#	plus 5 eighth notes.  Thus, the sixth row begins after a time equal
#	to 21 eighth notes.
#
#	The TEMPO is in units of beats per minute, where one beat is a
#	quarter note.  The duration symbols are
#
#	!				sixteenth
#	*				eighth
#	q				quarter
#	h				half
#	o				whole
#
#
#	Note Patterns
#	--------------
#
#	A note pattern is the basic unit given on a score line.  Patterns
#	are sequences of notes, and notes are specified by a list of variables
#	or literal numbers that form a Csound score line.  Consider the following
#	Csound score line:
#
#		i 8		0	0.6		0.9		7.02
#
#	where p1 is the instrument number, p2 is the start time, p3 is the
#	duration, p4 is the amplitude, and p5 is the pitch.  You would define this
#	line as a one-note pattern in Ascore with the following code:
#
#		set MyNote 		& 8 & q 0.9 7.02
#
#	or equivalently
#
#		set MyNote {
#			&
#			8
#			&
#			q
#			0.9
#			7.02
#		}
#
#	The "set" keyword is used to define a variable or list of variables.
#	In a note pattern definition the first element in the list is a local
#	tempo specifier.  In the above example, the tempo is given as "&" which
#	means use the global TEMPO.  The remaining elements in the list after
#	the tempo specifier have the same order and meaning as the Csound
#	score line.  The second elemnt is the instrument number.  The third
#	element is the start time.  Notice that in the example an "&" symbol
#	is used for the start time.  This means that the note starts at the
#	current score time (the sum of the duration symbols in all rows
#	preceding it), so you don't have to keep recalculating the time in
#	seconds.  The fourth element is the duration, and here we see the
#	relative duration symbol for a quarter note, again so that you don't
#	have to calculate time in seconds.  The fifth element in the list is
#	the amplitude, and the sixth element is the pitch.  These values are
#	given as literal numbers, but you can also specify them as variables.
#	Consider the following variable definitions:
#
#		set AmpNormal	0.9
#		set FirstPitch	7.02
#
#	Now the note pattern definition can be rewritten with these new
#	variables:
#
#		set MyNote		& 8 & q AmpNormal FirstPitch
#
#	Any other note pattern definition can now use these variables.
#
#
#	Start Time Symbols
#	-------------------
#
#	In the above examples we saw that the starting time of a note can
#	be specified as the current score time by using the "&" symbol.  This
#	"relation symbol", as I call it, is actually a simple linear function
#	that can have a gain and an offset.  Specifying "&" is equivalent to
#	writing
#
#		(1 * &)   +   0	
#
#	or, 1 times the current score time plus 0 seconds.  This way you can
#	make subtle modifications to the note starting time such as
#
#		0.95&
#		&+1.01
#		2.5&+0.001
#
#	The syntax of the relation symbols requires that the gain (multiplier)
#	be written with no multiplication symbol so that "2 times &" must be
#	written as  "2&" not as "2*&" (the "*" symbol is already used as a
#	duration symbol).  Also, there may not be any whitespace between the
#	gain, "&", and offset specifiers.
#
#	The offset to a relation symbol can be either a literal number in
#	units of seconds or a duration symbol.
#
#		&+[5*]
#
#	means the current score time plus 5 eighth notes.  The duration symbols
#	must be enclosed in square brackets in a start-time definition.
#
#
#	Two Different Start Times
#	--------------------------
#
#	There are two start time symbols:  "&" and "@".  The "&" in the above
#	examples is the current score time but also relative to the current note
#	pattern.  If the note pattern contained, say, two notes, then the "&"
#	in the second note's start time means "the current score time PLUS the
#	duration of all preceding notes in this pattern".  You can think of
#	"&" as a sequential start time for notes in a pattern.  If each note
#	in a pattern is defined with "&" as the start time, then the notes
#	follow one after the other in time.
#
#	The second start time is the "@" symbol.  It means the absolute score
#	time.  If all notes in a pattern are defined with "@" as the start time
#	then they all start simultaneously.  You can thing of "@" as a parallel
#	start time for notes in a pattern.  It would be used for defining a
#	piano chord, for example.  Also, if you want notes to overlap, then you
#	would use the "@" start time symbol with offsets:
#
#		@   @+[q]
#
#	which says that the first note starts at the absolute current score time
#	and the second note starts at the same time, but after a quarter note
#	duration.  If the first note was defined with a half note duration, then
#	the second note would start while the first note was still playing.
#
#	Global Tempo and Local Tempo
#	-----------------------------
#
#	The global tempo is given with the TEMPO statement and affects all
#	note durations and start times that occur after it is read.  This
#	means the score times that are on the far left of the scoreline section
#	and the start times of individual note in the note patterns.  A local
#	tempo can be defined within a note pattern.  The first field in a
#	note pattern is a local tempo, which is the exception to the
#	correspondence to csound p-fields:
#
#		&	^in_po_1  ^st_po_1   ^dur_po_1  ^amp_po_1  ^key_po_1
#
#	In the above note pattern, the & symbol is for the local tempo, and
#	the following fields are instrument number, start time, duration,
#	amplitude, and pitch, exactly in the same order as the pfields in the
#	resulting csound score.  The & symbol is not translated into a csound
#	pfield.  It is the local tempo used (notwithstanding the global tempo)
#	to calculate the start times and durations within this pattern.  In
#	this case, there is a & symbol, which means use the global tempo. A
#	local tempo field with 0.5& would mean use the global tempo divided
#	by 2.  Or an absolute number can be given for the local tempo.  Remember
#	that THE LOCAL TEMPO CANNOT CHANGE WHEN THE FOLLOWING PATTERN STARTS.
#	The start of each pattern is determined by the score-line  section of
#	the ascore script, and each pattern duration at the far left of all the
#	tracks is calculated only with the global tempo.  Therefore, to slow
#	down (or speed up) a succession of patterns, it is necessary to put
#	a different global tempo statement before each pattern.  Then, to obtain
#	slowing (or speeding) of notes within each pattern, use its local tempo.
#
#
#
#
#----------------------------------------------------------------------
#
#	Symbol Reference
#	------------------
#
#
#	&	in the start time column is the global start time PLUS the
#		durations of the previous notes in the local group: that is
#		it is the global start time plus the local durations
#		<gain>&+<offset>, for example 2.3&+3
#
#	@	in the start time column is the global start time
#		disregarding the previous nore durations in the local
#		group: this is useful for defining chords
#
#	^	if a list is to be evaluated sequentially in time, then
#		its variable name must be preceded by a "^"
#




BEGIN {

	tracks = 0
	tempo = 100
	scorelineno = 0
	scoretime = 0
	sectcount = 0
	varcount = 0
	reverbcount = 0
	skip_score = 0
	skip_track[1] = 0
	have_skiptrack = 0
	use_random = 0
	random_range = 0
	srand()

	#debug_child = 1	# DEBUGGER


	print "ASCORE ver. 3.5 (051009), copyright 2005, Alfred Steffens Jr."
	print "GNU General Public License, see the file called COPYING"
	#
	# defaults
	#
	evaluate_durations(tempo)
}

#
#	ignore comments
#
$1 ~ /#/ {
	next
}

#
#	ignore blank lines
#
NF == 0 {
	next
}

#
#	section statement (not used)
#
$1 ~ /SECTION/ {
	#
	#	A section definition
	#

	#
	#	must have at least 9 fields
	#
	#	1 = SECTION
	#	2 = name
	#	3 = awk file for riff or "sect"
	#	4 = variable name to hold instr #
	#	5 = variable name to hold start time
	#	6 = variable name to hold tempo
	#	7 = variable name to hold key
	#	8 = variable name to hold amplitude
	#	9 = variable name to hold pan
	#

	sectcount += 1
	sect_cmd[sectcount] = sprintf("echo junk | awk -f %s instr=$%s start=$%s\
			 tempo=$%s key=$%s amp=$%s pan=$%s >> hold.sco", \
			$3, $4, $5, $6, $7, $8, $9)
	sect_name[sectcount] = $2
	sect_file[sectcount] = $3
	sect_instr_name[sectcount] = $4
	sect_start_name[sectcount] = $5
	sect_tempo_name[sectcount] = $6
	sect_key_name[sectcount] = $7
	sect_amp_name[sectcount] = $8
	sect_pan_name[sectcount] = $9

	#print sect_name[sectcount]
	#print sect_cmd[sectcount]

	next

} # SECTION




#
#	tracks statement
#
$1 ~ /TRACKS/ {

	if (skip_score) {
		next
	}

	if (NF < 2) {
		print "ERROR: TRACKS statement syntax: need \"TRACKS <number>\" " | "cat 1>&2" 
		exit 1
	}

	tracks = $2
	print "TRACKS = " tracks
	next

} # TRACKS


#
#	tempo statement
#
$1 ~ /TEMPO/ {

	if (skip_score) {
		next
	}

	if (NF < 2) {
		print "ERROR: TEMPO statement syntax: need \"TEMPO <number>\" " | "cat 1>&2" 
		exit 1
	}

	tempo = $2
	evaluate_durations(tempo)
	next

} #  TEMPO


#
#	set statement
#
$1 ~ /set/ {

	if (skip_score) {
		next
	}

	if (NF < 3) {
		print "ERROR: set statement syntax: need \"set <var_name> <number>\" " | "cat 1>&2" 
		exit 1
	}
	fields_in_set_statement = NF

	#
	#	check if there is already a variable with that name
	#
	var_already_exists = 0
	for (i=1; i<varcount; i++) {
		if ($2 == var_name[i]) {
			var_already_exists = 1
			varindex = i
		}
	}

	if (var_already_exists) {
		# use the index found above
	}
	else {
		#
		# new variable, increment count
		#
		varcount += 1
		varindex = varcount
	}

	var_name[varindex] = $2

	#
	#	since this is not a function (but may be redefined
	#	later as one)
	#
	var_op[varindex] = "null"
	var_operand[varindex] = "null"
	var_reference[varindex] = "null"


	#
	#	possibilities for value:
	#	1.	a number  (NF == 3, arg3 is a number)
	#	2.	another variable  (NF==3, arg3 starts with "$")
	#	3.	a list  (NF==3, arg3 is a "{" )
	#
	if ($3 == "{") {
		#
		#	read lines until a closing "}" is found
		#
		var_size[varindex] = 0
		getline
		while ($1 != "}") {

			if (NF > 1) {
				print "ERROR: set statement syntax: 1 field allowed in list" | "cat 1>&2" 
				exit 1
			}

			if (is_an_alias($1)) {
				var_value[varindex] = sprintf("%s %s", var_value[varindex], insert_alias($1))
				var_size[varindex] += 1
			}
			else if (is_a_var_reference($1)) {
				j = get_var_index_from_reference($1)

				if (is_an_operation(var_name[j])) {
					newvalue = perform_operation(var_reference[j], var_op[j], var_operand[j])
					var_value[varindex] = sprintf("%s %s", \
						var_value[varindex], newvalue)
					var_size[varindex] += 1
					#print "set a VAR REFERENCE field " $3 " value= (" var_value[varindex] ")"
				}
				else {
					var_value[varindex] = sprintf("%s %s", \
						var_value[varindex], var_value[j])
					var_size[varindex] += var_size[j]
					#print "set a VAR REFERENCE field " $1
					#print "value= (" var_value[varcount] ")"
				} # else -- is not an operation
			}
			else {
				var_value[varindex] = sprintf("%s %s", var_value[varindex], $1)
				var_size[varindex] += 1
			}

			getline
		} # while

	}
	else {
		#
		# 	a single field or list specified on one line
		#
		# copy the string starting at field #3
		#
		#-----------------------------------------------
		#	the following code can give the wrong answer
		#	which is why it is not used
		#	ind = index($0, $3)
		#	var_value[varcount] = substr($0, ind)
		#-----------------------------------------------

		#
		#	first field (number 3)
		#
		if (is_an_alias($3)) {
			var_value[varindex] = insert_alias($3)
			var_size[varindex] += 1
		}
		else if (is_a_var_reference($3)) {

			i = get_var_index_from_reference($3)
			if (is_an_operation(var_name[i])) {
				#
				#	if this is an "operation" ---> from eval
				#
				var_value[varindex] = perform_operation(var_reference[i], var_op[i], var_operand[i])
				var_size[varindex] += 1
				#print "set a VAR REFERENCE field " $3 " value= (" var_value[varindex] ")"
			}
			else {
				#
				#	else just a normal variable value
				#
				var_value[varindex] = var_value[i]
				var_size[varindex] += var_size[i]
				#print "var_size= " var_size[varindex]
				#print "set a VAR REFERENCE field " $3
				#print "value= (" var_value[varcount] ")"
			}
		}
		else {
			var_value[varindex] = $3
			var_size[varindex] += 1
		}

		#
		#	fields 4 through the last
		#
		for (i=4; i<=NF; i++) {

			if (is_an_alias($i)) {
				var_value[varindex] = sprintf("%s %s", \
					var_value[varindex], insert_alias($i))
				var_size[varindex] += 1
			}
			else if (is_a_var_reference($i)) {
				j = get_var_index_from_reference($i)

				if (is_an_operation(var_name[j])) {
					newvalue = perform_operation(var_reference[j], var_op[j], var_operand[j])
					var_value[varindex] = sprintf("%s %s", \
						var_value[varindex], newvalue)
					var_size[varindex] += 1
					#print "set a VAR REFERENCE field " $3 " value= (" var_value[varindex] ")"
				}
				else {
					var_value[varindex] = sprintf("%s %s", \
						var_value[varindex], var_value[j])
					var_size[varindex] += var_size[j]
					#print "var_size= " var_size[varindex]
					#print "set a VAR REFERENCE field " $i
					#print "value= (" var_value[varcount] ")"
				} # else -- not an operation -- just a value
			}
			else {
				var_value[varindex] = sprintf("%s %s", \
					var_value[varindex], $i)
				var_size[varindex] += 1
			}

		} # for -- i

	} # else -- a single line

	#print "set: found " var_size[varindex] " fields"
	#print "set:  (" var_value[varindex] ") from (" $0 ") at index " ind

#	if (fields_in_set_statement == 3) {
#		a1 = Sc_EvalVar(var_name[varindex])
#		print "var " var_name[varindex] " evaluates to " a1
#		if (is_a_numeric_var(var_value[varindex])) {
#			print var_value[varindex] " is a number"
#		}
#	}
	next

} #  set



#
#	seteval statement
#
$1 ~ /eval/ {

	if (skip_score) {
		next
	}

	#
	#	eval should be used like the following examples:
	#
	#	eval newvar $oldvar * 3.4
	#	eval newvar $oldvar / 3.4
	#	eval newvar $oldvar + 3.4
	#	eval newvar $oldvar - 3.4
	#
	#	eval newvar $oldvar <op> <number>
	#
	#	or
	#
	#	eval newvar $oldvar1 <op> $oldvar2
	#

	if (NF < 5) {
		print "ERROR: seteval statement syntax: need \"seteval <var_name> <$var> op <number>\" "  | "cat 1>&2"
		exit 1
	}
	fields_in_set_statement = NF

	#
	#	check if there is already a variable with that name
	#
	var_already_exists = 0
	for (i=1; i<varcount; i++) {
		if ($2 == var_name[i]) {
			var_already_exists = 1
			varindex = i
		}
	}

	if (var_already_exists) {
		# use the index found above
	}
	else {
		#
		# new variable, increment count
		#
		varcount += 1
		varindex = varcount
		var_name[varindex] = $2
	}


	#
	#	the 3rd argument should be the reference to another
	#	variable:  $var1
	#
	if (is_a_var_reference($3)) {
		#
		#	we don't need the index yet, because we wont evaluate
		#	the function now
		#
		i = get_var_index_from_reference($3)
		var_reference[varindex] = $3
		var_value[varindex] = "null"
		var_size[varindex] = 1
	}
	else {
		print "ERROR: seteval syntax: 1st argument must be a" \
			" reference ($ prefix)" | "cat 1>&2" 
		exit 1
	}

	#
	#	for now we just store the operation
	#
	var_op[varindex] = $4
	var_operand[varindex] = $5

	#
	#	that's all for now--evaluation takes place at Sc_EvalVar
	#

	next

} # eval



$1 ~ /child/ {

#
#		child  newname  oldname  column_to_mod  operator  arg1  arg2 ....
#

	if (skip_score) {
		next
	}

	if (NF < 6) {
		print "ERROR: child statement syntax: not enough arguments for line (" $0 ")" | "cat 1>&2" 
		exit 1
	}

	#
	#	create the "argstring" to be passed to Mod_ScVar
	#

	#
	#	allow comments at the end of the "child" line
	#
	comchd_start = 0
	for (i=1; i<=NF; i++) {
		if (index($i, "#") == 1) {
			#	found a comment on the child line
			comchd_start = i
			break
		}
	}

	if (comchd_start) {
		#
		#	comments were found at the end
		#
		modscvar_argstring = sprintf("%s", $4)
		for (i=5; i<=(comchd_start-1); i++) {
			modscvar_argstring = sprintf("%s %s", modscvar_argstring, $i)
		}
	}
	else {
		#
		#	no comments
		#
		modscvar_argstring = sprintf("%s", $4)
		for (i=5; i<=NF; i++) {
			modscvar_argstring = sprintf("%s %s", modscvar_argstring, $i)
		}
	}


	if (debug_child) {
	print "DEBUGGER: child: argstring= (" modscvar_argstring ")"
	}

	#
	#	arguments:
	#
	#		newname,  oldname,  argstring
	#
	#	where "argstring" is $4 $5 .....$NF
	#
	Mod_ScVar($2, $3, modscvar_argstring)

	#
	#	now the above call to Mod_ScVar will have added a new variable to the
	#	table with the name of "newname"
	#

	if (debug_child) {
	#	DEBUGGER : find the new variable
	i = get_var_index_from_name($2)
	print "DEBUGGER: child: new variable has index (" i ")"
	}


	next
} #  child


#
#	alias	definition
#
$1 ~ /alias/ {
	#
	#	usage:
	#
	#	alias <alias_name> {
	#		<input_1> <output_1>
	#		<input_2> <output_2>
	#		<input_3> <output_3>
	#			.		.
	#			.		.
	#		<input_N> <output_N>
	#
	#	example definition:
	#
	#	alias noteinstr1 {
	#		7.05	20
	#		7.06	20
	#		7.07	21
	#		7.08	21
	#	}
	#
	#	example in the score:
	#
	#	~noteinstr1(7.05)
	#	will be substituted with 20
	#	~noteinstr1(7.06)
	#	will be substituted with 20
	#	~noteinstr1(7.07)
	#	will be substituted with 21
	#	~noteinstr1(7.08)
	#	will be substituted with 21
	#
	#
	#	code:
	#	alias_list[name, input] = output
	#

	curr_alias_name = $2
	if ($3 != "{") {
		print "ERROR: alias statement syntax: curly bracket not found" | "cat 1>&2" 
		exit 1
	}

	#
	#	read the following lines until a closing brace is found
	#
	getline
	while ($1 != "}") {

		if (NF < 2) {
			print "ERROR: alias statement syntax: at least 2 fields needed on a line" | "cat 1>&2" 
			exit 1
		}

		alias_list[curr_alias_name, $1] = $2

		getline
	} # while

	next
} # alias


#
#	Header file
#
$1 ~ /HEADER/ {

	if (skip_score) {
		next
	}

	if (NF < 2) {
		print "ERROR: HEADER statement syntax: need \"HEADER <filename>\" " | "cat 1>&2" 
		exit 1
	}

	cmd = sprintf("cat %s > hold.sco", $2)
	system(cmd)
	next

} #  HEADER



#
#	Include file
#
$1 ~ /INCLUDE/ {

	if (skip_score) {
		next
	}

	if (NF < 2) {
		print "ERROR: INCLUDE statement syntax: need \"INCLUDE <filename>\" " | "cat 1>&2" 
		exit 1
	}

	cmd = sprintf("cat %s >> hold.sco", $2)
	system(cmd)
	next

} #  INCLUDE




#
#	End
#
$1 ~ /END/ {

	if (skip_score) {
		next
	}

	print ";----------------------------------------------------------------" >> "hold.sco"
	print ";	End of Score" >> "hold.sco"
	print "e" >> "hold.sco"
	print ";----------------------------------------------------------------" >> "hold.sco"
	print "" >> "hold.sco"
	next
} #   END



#
#	REVERB macros
#
$1 ~ /REVERB/ {

	if (skip_score) {
		next
	}

	#
	#	must have at least 4 fields
	#
	#	1 = REVERB
	#	2 = name
	#	3 = awkfile
	#	4 = first instrument number
	#
	if (NF < 4) {
		print "ERROR: REVERB statement syntax: need \"REVERB <var_name> <awkfile> <instr1> <instr2> ...\" " | "cat 1>&2" 
		exit 1
	}

	reverbcount += 1
	reverb_name[reverbcount] = $2
	reverb_file[reverbcount] = $3
	#
	#	allow comments at the end of a REVERB line
	#
	comrvb_start = 0
	for (i=1; i<=NF; i++) {
		if (index($i, "#") == 1) {
			#	found a comment on the REVERB line
			comrvb_start = i
			break
		}
	}
	if (comrvb_start) {
		#
		#	comments were found at the end of the REVERB line
		#	starting at field number comrvb_start
		#
		reverb_numinstr[reverbcount] = (comrvb_start-1) - 3
	}
	else {
		#
		#	no comments
		#
		reverb_numinstr[reverbcount] = NF - 3
	}
	numi = reverb_numinstr[reverbcount]
	#
	#	all the instrument numbers form a string
	#
	if (comrvb_start) {
		#
		#	get the remainder of the line -- but leave off the comment
		#
		cindex = index($0, "#")
		reverb_instr[reverbcount] = substr($0, match($0, /\ [1-9][0-9]*/)+1, (cindex-1))
	}
	else {
		#
		#	get the remainder of the line
		#
		reverb_instr[reverbcount] = substr($0, match($0, /\ [1-9][0-9]*/)+1)
	}
	#print "reverb instr string: " reverb_instr[reverbcount]

	next

} #		REVERB


#
#	turn Reverb On
#
$1 ~ /RVB_ON/ {

	if (skip_score) {
		next
	}

	if (NF < 3) {
		print "ERROR: RVB_ON statement syntax: need \"RVB_ON <name> <amp>\" " | "cat 1>&2" 
		exit 1
	}

	#
	# find the index corresponding to the Reverb name
	#
	found_reverb=0
	for (i=1; i<=reverbcount; i++) {

		if ($2 == reverb_name[i]) {
			reverb_index = i
			found_reverb=1
			break;
		} # if

	} # for i

	if (found_reverb) {
		reverb_start[reverb_index] = scoretime
		reverb_amp[reverb_index] = $3
	}

	next
} #  RVB_ON


#
#	turn Reverb Off
#
$1 ~ /RVB_OFF/ {

	if (skip_score) {
		next
	}

	if (NF < 2) {
		print "ERROR: RVB_OFF statement syntax: need \"RVB_OFF <name>\" " | "cat 1>&2" 
		exit 1
	}

	#
	# find the index corresponding to the Reverb name
	#
	found_reverb=0
	for (i=1; i<=reverbcount; i++) {

		if ($2 == reverb_name[i]) {
			reverb_index = i
			found_reverb=1
			break;
		} # if

	} # for i

	if (found_reverb) {

		reverb_end[reverb_index] = scoretime
		rvb_duration = reverb_end[reverb_index] - reverb_start[reverb_index]

		#
		#	get individual instr numbers
		#
		split(reverb_instr[reverb_index], rvb_ary)
		for (i=1; i<=reverb_numinstr[reverb_index]; i++) {

			instr = rvb_ary[i]
			cmd = sprintf("echo junk | awk -f %s instr=%d start=%g duration=%g amp=%g >> hold.sco", \
					reverb_file[reverb_index], 
					instr, reverb_start[reverb_index], 
					rvb_duration, reverb_amp[reverb_index])

			#print "reverb awk cmd = (" cmd ")"
			system(cmd)
		} # for i
	}

	next
} #  RVB_OFF


#
#	Use SKIPSCORE AND STARTSCORE together
#
$1 ~ /SKIPSCORE/ {

	#
	#	signal to skip all following score lines
	#	until a STARTSCORE statement is found
	#

	skip_score = 1
	next

} #  SKIPSCORE

#
#	Use SKIPSCORE AND STARTSCORE together
#
$1 ~ /STARTSCORE/ {

	#
	#	turn score lines back on
	#

	skip_score = 0
	next

} #  SKIPSCORE

#
#	SKIPTRACK
#	usage:
#		SKIPTRACK 2 3 6
#	will cause tracks 2, 3, and 6 to be skipped until a
#	SKIPTRACK with no arguments is found, which resets the skip
#	to "don't skip"
#
$1 ~ /SKIPTRACK/ {
	#
	#	expect a white-space separated list of tracks to skip
	#

	#
	#	first, clear any previous SKIPTRACK statement
	#
	for (j=1; j<=tracks; j++) {
		if ((j) in skip_track) {
			delete skip_track[j]
		}
	}
	skip_track[1] = 0
	have_skiptrack = 0

	#
	#	set the tracks to skip
	#
	j = 1
	for (i=2; i<=NF; i++) {

		itracknumber = $i
		if (is_a_number(itracknumber)) {

			skip_track[j] = itracknumber
			j++

			have_skiptrack = 1
		} # if -- is_a_number
	} # for -- i

	next

} #  SKIPTRACK



#
#	RANDOMIZE
#	usage:
#		RANDOMIZE 0.X
#	Adds a random offset (uniform prob dist=white noise) to
#	every start time after this statement.  Using a RANDOMIZE
#	statement with no argument turns it off.  The argument
#	should be a fraction between 0 and 1 which is the percent
#	of a quarter.
#
#	example:
#		RANDOMIZE 0.21
#	means that at most +- 21 percent of a quarter note will be
#	added (subtracted) from the start time.  The seed is given at
#	the start of the score.
#
$1 ~ /RANDOMIZE/ {

	#
	#	reset any previous RANDOMIZE statement
	#
	use_random = 0
	random_range = 0

	#
	#	look for the range argument
	#
	if (NF >= 2) {
		if (is_a_number($2)) {
			random_range = $2
			use_random = 1
		}
	}

	next

} # RANDOMIZE






#	/////////////////////////////////////////////////////////////////
#	score line
#	Default line: a score line
#	No comments allowed on this line
#	/////////////////////////////////////////////////////////////////
#
{

	#
	#	if the "skip_score" field is set, this is like commenting-out
	#	the following score lines
	#
	if (skip_score) {
		next
	}



	scorelineno += 1

	if (NF != (tracks+1)) {

		print "ERROR: score line " scorelineno ": number of columns NOT EQUAL value of TRACKS" | "cat 1>&2" 
		print "line: (" $0 ")"
		exit 1

	} # -- if NF


	print "timing: " $1
	#
	#	read the time field
	#	if the field is a number, then the start time taks this value
	#
	if (is_a_number($1)) {
		print "score time set " $1
		scoretime = $1
	}
	else {
		print "score time  " scoretime
	}



	#
	#	loop through the track fields, see if any are section names
	#
	for (i=2; i<=NF; i++) {

		#
		#	July 4, 2005 -- SKIPTRACK
		#
		if (have_skiptrack) {
			#
			#	if the "SKIPTRACK" keyword is found
			#	we must skip that track in the score section
			#
			if (match_skiptrack(i-1)) {
				#
				#	ignore this field (column) and continue to
				#	the next column
				#
				continue
			}
		} # if -- skip_track


		found_sect = 0
		for (j=1; j<=sectcount; j++) {
			#
			#	sect name ?
			#
			if ($i == sect_name[j]) {
				found_sect = 1
				sect_num = j
				break;
			} # if -- matched a section name

		} #  for j

		#
		#	if we found a section name, 
		#	then look for its variables
		#
		if (found_sect) {

			print "sect : " sect_name[sect_num] " : line " scorelineno " : field " i

			#
			#	find instr name
			#
			found_instr=0
			for (k=1; k<=varcount; k++) {

				if (var_name[k] == sect_instr_name[sect_num]) {
					instr = var_value[k]
					found_instr=1
				} # match instr name

			} # for k

			if (found_instr==0) {
				print "ERROR: no instr variable: scoreline " scorelineno ", sect " sect_name[sect_num] | "cat 1>&2" 
				exit 1
			}


			#
			#	find start name
			#
			found_start=0
			for (k=1; k<=varcount; k++) {

				if (var_name[k] == sect_start_name[sect_num]) {
					start = var_value[k]
					found_start=1
				} # match start name

			} # for k

			if (found_start==0) {
				#
				#	use the current score time
				#
				start = scoretime
			}

			#
			#	find tempo name
			#
			found_tempo=0
			for (k=1; k<=varcount; k++) {

				if (var_name[k] == sect_tempo_name[sect_num]) {
					ltempo = var_value[k]
					found_tempo=1
				} # match tempo name

			} # for k

			if (found_tempo==0) {
				#
				# just use the global tempo
				#
				ltempo = tempo
			}


			#
			#	find key name
			#
			found_key=0
			for (k=1; k<=varcount; k++) {

				if (var_name[k] == sect_key_name[sect_num]) {
					key = var_value[k]
					found_key=1
				} # match key name

			} # for k

			if (found_key==0) {
				print "ERROR: no key variable: scoreline " scorelineno ", sect " sect_name[sect_num] | "cat 1>&2" 
				exit 1
			}

			#
			#	find amp name
			#
			found_amp=0
			for (k=1; k<=varcount; k++) {

				if (var_name[k] == sect_amp_name[sect_num]) {
					lamp = var_value[k]
					found_amp=1
				} # match amp name

			} # for k

			if (found_amp==0) {
				print "ERROR: no amplitude variable: scoreline " scorelineno ", sect " sect_name[sect_num] | "cat 1>&2" 
				exit 1
			}


			#
			#	find pan name
			#
			found_pan=0
			for (k=1; k<=varcount; k++) {

				if (var_name[k] == sect_pan_name[sect_num]) {
					pan = var_value[k]
					found_pan=1
				} # match pan name

			} # for k

			if (found_pan==0) {
				pan = 0.5
			}

			sect_cmd[sect_num] = sprintf("echo junk | awk -f %s instr=%d start=%g tempo=%g key=%.2f amp=%g pan=%g >> hold.sco", \
					sect_file[sect_num], instr, start, ltempo, key, lamp, pan)

			#print "found command (" sect_cmd[sect_num] ")"
			system(sect_cmd[sect_num])

			continue

		} # if -- found_sect


		if (is_in_varlist($i)) {
			vret = Sc_EvalVar($i)
			#print "DEBUGGER: " $i " into Sc_EvalVar returned:"
			#print_string_matrix(vret)
			#
			#	output the variable expansion to the csound score
			#
			string_matrix_to_csound(vret)
		}


	} # for -- fields on track line



	#
	# read the time field (last thing to do)
	#
	if (is_a_number($1)) {
	}
	else {
		scoretime += get_duration($1)
		print "time inc to " scoretime
	}

} #  score line



END {


}


#	----------------------------------------------------------------------
#	----------------------------------------------------------------------
#	----------------------------------------------------------------------
#	----------------------------------------------------------------------
#	----------------------------------------------------------------------
#	----------------------------------------------------------------------

function evaluate_durations(tempo) {
#
#	produces global values
#

	qdur = (1/tempo) * 60	# quarter note 	( "q" in script )
	print "q = " qdur
	edur = qdur/2			# eighth note	( "*" in script )
	print "* = " edur
	sdur = edur/2			# sixteenth		( "!" in script )
	print "! = " sdur
	
	hdur = qdur * 2			# half note		( "h" in script )
	print "h = " hdur
	wdur = qdur * 4			# whole note	( "o" in script )
	print "o = " wdur
}
#--------------------- evaluate_durations -------------------------------



function evaluate_qdur(ltempo,   dur) {
#
#	produces local value
#

	dur = (1/ltempo) * 60	# quarter note 	( "q" in script )
	return dur
}
#--------------------- evaluate_qdur -------------------------------



function evaluate_hdur(ltempo,   dur) {
#
#	produces local value
#

	dur = (1/ltempo) * 60	# quarter note 	( "q" in script )
	dur = dur*2				# half note		( "h" in script )
	return dur
}
#--------------------- evaluate_hdur -------------------------------



function evaluate_wdur(ltempo,   dur) {
#
#	produces local value
#

	dur = (1/ltempo) * 60	# quarter note 	( "q" in script )
	dur = dur*4				# whole note	( "o" in script )
	return dur
}
#--------------------- evaluate_wdur -------------------------------



function evaluate_edur(ltempo,   dur) {
#
#	produces local value
#

	dur = (1/ltempo) * 60	# quarter note 	( "q" in script )
	dur = dur/2				# eighth note	( "*" in script )
	return dur
}
#--------------------- evaluate_edur -------------------------------



function evaluate_sdur(ltempo,   dur) {
#
#	produces local value
#

	dur = (1/ltempo) * 60	# quarter note 		( "q" in script )
	dur = dur/4				# sixteenth note	( "!" in script )
	return dur
}
#--------------------- evaluate_sdur -------------------------------



function is_a_numeric_var(field) {
#
#	used on a "set" command
#
	if (match(field, /\$/)) {
		return 0
	}
	if (match(field, /[a-zA-Z]/)) {
		return 0
	}
	if (match(field, /[\&\*\!\@]/)) {
		return 0
	}
	if (match(field, /[0-9][0-9]*(\.)*[0-9]*/)) {
		return 1
	}
	return 0
}
# --------------------- is_a_numeric_var ---------------------------------



function is_a_relation_symbol(field) {
#
#	"field" is a value, not a name, similar to "is_a_numeric_var()"
#

	if (match(field, /\^./)) {
		return 0
	}
	else if (field == "c") {
		return 1
	}
	else if (match(field,/([0-9][0-9]*(\.)*[0-9]*)*o(\+[0-9][0-9]*(\.)*[0-9]*)*/)) {
		#
		#	whole note
		#	decimal number times a duration
		#
		return 1
	}
	else if (match(field,/([0-9][0-9]*(\.)*[0-9]*)*h(\+[0-9][0-9]*(\.)*[0-9]*)*/)) {
		#
		#	half note
		#	decimal number times a duration
		#
		return 1
	}
	else if (match(field,/([0-9][0-9]*(\.)*[0-9]*)*q(\+[0-9][0-9]*(\.)*[0-9]*)*/)) {
		#
		#	quarter note
		#	decimal number times a duration
		#
		return 1
	}
	else if (match(field,/([0-9][0-9]*(\.)*[0-9]*)*\*(\+[0-9][0-9]*(\.)*[0-9]*)*/)) {
		#
		#	eighth note
		#	decimal number times a duration
		#
		return 1
	}
	else if (match(field,/([0-9][0-9]*(\.)*[0-9]*)*\!(\+[0-9][0-9]*(\.)*[0-9]*)*/)) {
		#
		#	sixteenth note
		#	decimal number times a duration
		#
		return 1
	}
	else if (match(field,/([0-9][0-9]*(\.)*[0-9]*)*\&(\+[0-9][0-9]*(\.)*[0-9]*)*/)) {  # scoretime offset by a number
		return 1
	}
	else if (match(field,/([0-9][0-9]*(\.)*[0-9]*)*\@(\+[0-9][0-9]*(\.)*[0-9]*)*/)) {  # scoretime offset by a number
		return 1
	}
	else {
		return 0;
	}

}
#-------------------- is_a_relation_symbol -------------------------------



function is_a_relation_scoretime(field) {
#
#	"field" contains "&"
#

	if (match(field,/([0-9][0-9]*(\.)*[0-9]*)*\&(\+[0-9][0-9]*(\.)*[0-9]*)*/)) {  # scoretime offset by a number
		return 1
	}
	else {
		return 0;
	}

}
#-------------------- is_a_relation_scoretime -------------------------------



function is_a_relation_ascoretime(field) {
#
#	"field" contains "@"
#

	if (match(field,/([0-9][0-9]*(\.)*[0-9]*)*\@(\+[0-9][0-9]*(\.)*[0-9]*)*/)) {  # scoretime offset by a number
		return 1
	}
	else {
		return 0;
	}

}
#-------------------- is_a_relation_ascoretime -------------------------------




function is_a_var_reference(field)
{
#
#	used on a "set" command
#

	if (match(field, /\$[a-zA-Z][a-zA-Z0-9]*/)) {
		return 1
	}
	return 0
}
#---------------------- is_a_var_reference ------------------------------



function get_var_index_from_name(field,   i) {
#
#	gets the var name index from variable "name" that was given
#	by name (no "$")
#

	for (i=1; i<=varcount; i++) {

		if (field == var_name[i]) {

			return i
		}

	} # for -- i

	return 0
}
#------------------ get_var_index_from_name -----------------------


function get_var_index_from_reference(field,   i,name) {
#
#	gets the var name index from variable "name" that was given
#	by reference as "$name"
#

	#
	#	assumes that the first character is a "$"
	#
	name = substr(field, 2)

	return get_var_index_from_name(name)
}
#------------------ get_var_index_from_reference -----------------------


function get_var_value(field)
{

	return val_value[get_var_index_from_name(field)]
}
#-------------------- get_var_value -------------------------------------



function is_in_varlist(field,    i,sfield)
{
	#
	#	ignores a leading "^" sign
	#

	for (i=1; i<=varcount; i++) {

		if (match(field, /\^/)) {
			#
			# take off the leading "^"
			#
			sfield = substr(field,2)
		} # if -- match
		else {
			sfield = field
		}

		if (sfield == var_name[i]) {
			return 1
		}
	}
	return 0
}
#--------------------- is_in_varlist ------------------------------------



function is_a_number(field) {
#
#	used on the score line
#


	if (match(field, /[0-9]*q/)) {
		return 0
	}
	else if (match(field, /[0-9]*\*/)) {
		return 0
	}
	else if (match(field, /[0-9]*h/)) {
		return 0
	}
	else if (match(field, /[0-9]*o/)) {
		return 0
	}
	else if (match(field, /[0-9]*!/)) {
		return 0
	}
	else if (match(field, /[0-9][0-9]*(\.[0-9]*)*/)) {
		return 1
	}
	else {
		return 0
	}
}
#---------------------- is_a_number --------------------------------------



function is_valid_op(opfield) {

	if (opfield == "*" || opfield == "/" || opfield == "+" || opfield == "-") {
		return 1
	}
	return 0
}
#---------------------- is_valid_op --------------------------------------



function is_an_operation(name,   i) {

	i = get_var_index_from_name(name)
	if (var_value[i] == "null" && is_valid_op(var_op[i])) {
		return 1
	}
	return 0
}
#---------------------- is_an_operation ----------------------------------



function perform_operation(reference, opfield, operand,   i,j,val,object) {

	#
	#	get the index for the variable in reference
	#	"reference" should be a $ followed by a variable name
	#
	i = get_var_index_from_reference(reference)
	if (is_an_operation(var_name[i])) {
		#
		#	the value in the reference must itself be evaluated
		#
		val = perform_operation(var_reference[i],var_op[i],var_operand[i])
	}
	else if (is_a_var_reference(var_name[i])) {
		#
		#	but, if the name is itself a variable reference
		#
		j = get_var_index_from_reference(var_name[i])
		val = var_value[j]
	}
	else {
		#
		# default -- the variable stored a number
		#
		if (is_a_numeric_var(var_value[i])) {
			val = var_value[i]
		}
		else {
			print "ERROR: perform_operation: can't interpret value " var_value[i] | "cat 1>&2" 
		}
	}


	#
	#	get the index of the value in the operand
	#
	if (is_a_numeric_var(operand)) {
		object = operand
	}
	else {
		#
		#	operand holds a variable name or reference
		#
		if (is_a_var_reference(operand)) {
			j = get_var_index_from_reference(operand)
		}
		else {
			print "ERROR: perform_operation: variable operand must be preceded by $" | "cat 1>&2" 
		}

		if (is_an_operation(var_name[j])) {
			#
			#	the value in the reference must itself be evaluated
			#
			object = perform_operation(var_reference[j],var_op[j],var_operand[j])
		}
		else if (is_a_var_reference(var_name[j])) {
			#
			#	but, if the name is itself a variable reference
			#
			j = get_var_index_from_reference(var_name[j])
			object = var_value[j]
		}
		else {
			#
			# default -- the variable stored a number
			#
			if (is_a_numeric_var(var_value[j])) {
				object = var_value[j]
			}
			else {
				print "ERROR: perform_operation: can't interpret value " var_value[j] | "cat 1>&2" 
			}
		}


	} # else -- operand is a variable



	#
	#	perform the operation
	#
	if (opfield == "*") {
		return (val * object)
	}
	else if (opfield == "+") {
		return (val + object)
	}
	else if (opfield == "/") {
		return (val / object)
	}
	else if (opfield == "-") {
		return (val - object)
	}

	#
	#
	# if we drop down to here there is something wrong
	#
	print "ERROR: perform_operation: " reference " " opfield " " operand | "cat 1>&2" 
	return "null"
}
#---------------------- is_valid_op --------------------------------------



function get_duration_gain(field,  ind,ind2,len,num)
{
#
#	in the relation, "a&+b"
#	returns "a"
#

	if (field == "!") {
		return 1;
	}
	else if (field == "*") {
		return 1;
	}
	else if (field == "q") {
		return 1;
	}
	else if (field == "h") {
		return 1;
	}
	else if (field == "o") {
		return 1;
	}

	ind = match(field, /[0-9][0-9]*(\.)*[0-9]*/)
	ind2 = match(field, /[\!\*qho]/)

	if (ind < ind2) {
		len = ind2 - ind
		num = substr(field, ind, len)
		return num
	}

	return 1
}
#---------------------- get_duration_gain -------------------------------



function get_duration_offset(field,   ind,ind2,sfd,off) {
#
#	in the relation, "a&+b"
#	returns "b"
#

	if (field == "!") {
		return 0;
	}
	else if (field == "*") {
		return 0;
	}
	else if (field == "q") {
		return 0;
	}
	else if (field == "h") {
		return 0;
	}
	else if (field == "o") {
		return 0;
	}


	ind = match(field, /\+/)
	if (ind) {
		sfd = substr(field, ind)
		ind2 = match(sfd, /[0-9][0-9]*(\.)*[0-9]*/)
		off = substr(sfd,ind2)	# skip past the "!*qho"
		return off
	}

	ind = match(field, /\-/)
	if (ind) {
		sfd = substr(field, ind)
		ind2 = match(sfd, /[0-9][0-9]*(\.)*[0-9]*/)
		off = substr(sfd,ind2)	# skip past the "!*qho"
		off *= -1
		return off
	}

	return 0
}
#---------------------- get_duration_offset -----------------------------



function get_duration(field,   ind,ind2,len,num,sfd,off) {


	ind = match(field, /\!/)
	if (ind) {
		num = get_duration_gain(field)
		off = get_duration_offset(field)
		return (num*sdur)+off
	}

	ind = match(field, /\*/)
	if (ind) {
		num = get_duration_gain(field)
		off = get_duration_offset(field)
		return (num*edur)+off
	}

	ind = match(field, /q/)
	if (ind) {
		num = get_duration_gain(field)
		off = get_duration_offset(field)
		return (num*qdur)+off
	}

	ind = match(field, /h/)
	if (ind) {
		num = get_duration_gain(field)
		off = get_duration_offset(field)
		return (num*hdur)+off
	}

	ind = match(field, /o/)
	if (ind) {
		num = get_duration_gain(field)
		off = get_duration_offset(field)
		return (num*wdur)+off
	}
	return -1
}
#---------------------- get_duration --------------------------------------




function get_duration_with_ltempo(field,ltempo,   ind,ind2,len,num,sfd,off,dur) {

	if (match(field,/([0-9][0-9]*(\.)*[0-9]*)*q(\+[0-9][0-9]*(\.)*[0-9]*)*/)) {

		dur = evaluate_qdur(ltempo)

		ind = match(field, /[0-9][0-9]*(\.)*[0-9]*/)
		if (ind) {
			ind2 = match(field, /q/)
			len = ind2 - ind
			num = substr(field, ind, len)
			sfd = substr(field,ind2+1)
			ind = match(sfd, /[0-9][0-9]*(\.)*[0-9]*/)
			if (ind) {
				off = substr(sfd, ind)
				return (num*dur)+off
			}
			return num*dur
		}

		return dur
	}
	else if (match(field,/([0-9][0-9]*(\.)*[0-9]*)*\*(\+[0-9][0-9]*(\.)*[0-9]*)*/)) {

		dur = evaluate_edur(ltempo)

		ind = match(field, /[0-9][0-9]*(\.)*[0-9]*/)
		if (ind) {
			ind2 = match(field, /\*/)
			len = ind2 - ind
			num = substr(field, ind, len)
			sfd = substr(field,ind2+1)
			ind = match(sfd, /[0-9][0-9]*(\.)*[0-9]*/)
			if (ind) {
				off = substr(sfd, ind)
				return (num*dur)+off
			}
			return num*dur
		}

		return dur
	}
	else if (match(field,/([0-9][0-9]*(\.)*[0-9]*)*h(\+[0-9][0-9]*(\.)*[0-9]*)*/)) {

		dur = evaluate_hdur(ltempo)

		ind = match(field, /[0-9][0-9]*(\.)*[0-9]*/)
		if (ind) {
			ind2 = match(field, /h/)
			len = ind2 - ind
			num = substr(field, ind, len)
			sfd = substr(field,ind2+1)
			ind = match(sfd, /[0-9][0-9]*(\.)*[0-9]*/)
			if (ind) {
				off = substr(sfd, ind)
				return (num*dur)+off
			}
			return num*dur
		}

		return dur
	}
	else if (match(field,/([0-9][0-9]*(\.)*[0-9]*)*o(\+[0-9][0-9]*(\.)*[0-9]*)*/)) {

		dur = evaluate_wdur(ltempo)

		ind = match(field, /[0-9][0-9]*(\.)*[0-9]*/)
		if (ind) {
			ind2 = match(field, /o/)
			len = ind2 - ind
			num = substr(field, ind, len)
			sfd = substr(field,ind2+1)
			ind = match(sfd, /[0-9][0-9]*(\.)*[0-9]*/)
			if (ind) {
				off = substr(sfd, ind)
				return (num*dur)+off
			}
			return num*dur
		}

		return dur
	}
	else if (match(field,/([0-9][0-9]*(\.)*[0-9]*)*!(\+[0-9][0-9]*(\.)*[0-9]*)*/)) {

		dur = evaluate_sdur(ltempo)

		ind = match(field, /[0-9][0-9]*(\.)*[0-9]*/)
		if (ind) {
			ind2 = match(field, /!/)
			len = ind2 - ind
			num = substr(field, ind, len)
			sfd = substr(field,ind2+1)
			ind = match(sfd, /[0-9][0-9]*(\.)*[0-9]*/)
			if (ind) {
				off = substr(sfd, ind)
				return (num*dur)+off
			}
			return num*dur
		}

		return dur
	}
	else {
		return -1	# error
	}
}
#---------------------- get_duration_with_ltempo ---------------------------------



function get_relation_gain(field,  ind,ind2,len,num)
{
#
#	in the relation, "a&+b"
#	returns "a"
#

	if (field == "&") {
		return 1;
	}
	if (field == "@") {
		return 1;
	}

	if (match(field, /([0-9\.]+)([&@])([+\-0-9\.\[\]!*qho])*/)) {
		ind = match(field, /[0-9][0-9]*(\.)*[0-9]*/)
		ind2 = match(field, /[\&\@]/)
	
		if (ind < ind2) {
			len = ind2 - ind
			num = substr(field, ind, len)
			return num
		}
		else {
			return 1
		}
	}
	else if (match(field, /([&@])([+\-0-9\.\[\]!*qho])*/)) {
		return 1;
	}

	return 1
}
#---------------------- get_relation_gain -------------------------------



function get_relation_offset(field,   ind,ind2,ind3,ind4,sfd,sfd2,len,off) {
#
#	in the relation, "a&+b"
#	returns "b"
#

	if (field == "&") {
		return 0;
	}
	if (field == "@") {
		return 0;
	}

	if (match(field, /([0-9\.]*)([&@])([+\-0-9\.\[\]!*qho])+/)) {

		ind = match(field, /[\+\-]/)
		sfd = substr(field, ind)
		if (ind) {
			#
			#	is offset a duration? ( for example, @+[7*] )
			#
			ind3 = match(sfd, /\[/)
			if (ind3) {
				ind3 = ind3+1
				ind4 = match(sfd,/\]/)
				len = ind4 - ind3
				sfd2 = substr(sfd,ind3,len)
				off = get_duration(sfd2)
			}
			else {
				#
				#	offset is just a number
				#
				ind2 = match(sfd, /[0-9][0-9]*(\.)*[0-9]*/)
				off = substr(sfd,ind2)	# skip past the "&"
			}
			return off
		}

	}
	else {
		return 0
	}

	return 0
}
#---------------------- get_relation_offset -----------------------------



function get_relation(field,currvalue,    gain,off,ret)
{
#
#	takes a "relation" expression such as 
#
#		&					(current value)
#		2.1&				(current value times a factor)
#		&+1.7				(current value plus an offset)
#		3&+20.2				(both)
#
#	and evaluates it like a function (similar to get_duration)
#	"currvalue" is a number to be substituted into "&", such
#	as a score time or a tempo
#
#	


	if (match(field,/([0-9][0-9]*(\.)*[0-9]*)*([\&\@])(\+[0-9][0-9]*(\.)*[0-9]*)*/)) {  # scoretime offset by a number

		gain = get_relation_gain(field)
		off = get_relation_offset(field)
		ret = (gain * currvalue + off)
		return ret
	}

	return "null"
}
# --------------------- get_relation --------------------------------------


function is_a_AND_relation(rel)
{
#
#	new with ver 2.9
#	if this is a relation of type "&" then it returns 1, else 0
#

	if (index(rel, "&")) {
		return 1
	}
	return 0
}
#-------------------- is_a_AND_relation -----------------------------------


function is_ANY_relation(rel)
{
#
#	new with ver 2.9
#	if this is a relation of type "&" or "@"  then it returns 1, else 0
#

	if (index(rel, "&")) {
		return 1
	}
	else if (index(rel, "@")) {
		return 1
	}
	return 0
}
#-------------------- is_ANY_relation -----------------------------------



function is_a_AT_relation(rel)
{
#
#	new with ver 2.9
#	if this is a relation of type "@" then it returns 1, else 0
#

	if (index(rel, "@")) {
		return 1
	}
	return 0
}
#-------------------- is_a_AT_relation -----------------------------------



function is_a_duration(rel,  i1,i2)
{
#
#	new with ver 2.9
#	if this is a relation of type "ohq*!" then it returns 1, else 0
#
#	because a duration symbol might occur in the offset of any kind
#	of relation, we must determine unambiguaously whether or not
#	this is a duration.
#

	#
	#	it can't be duration is any of these two symbols are present
	#
	if (index(rel, "&")) {
		return 0
	}
	if (index(rel, "@")) {
		return 0
	}

	i1 = match(rel, /ohq*!/)
	if (i1 == 0) {
		return 0
	}


	#
	#	check if there are any "[  ]" constructions
	#
	i2 = index(rel, "[")
	if (i2 > 0) {

		if (i1 < i2) {
			return 1
		}
		else {
			return 0
		}

	}
	else {

		return 1

	}

	return 0
}
#-------------------- is_a_duration -----------------------------------



function parse_relation_or_duration_string(rel,   i,j,len,srel,urel,sparse,result,a1,a2,a3,a4,a5,a6,a7)
{
#
#	takes any relation string or duration string of arbitrary format:
#	<a> & +- <b>
#	<a> @ +- <b>
#	<a> o +- <b>
#	<a> h +- <b>
#	<a> q +- <b>
#	<a> * +- <b>
#	<a> ! +- <b>
#
#	where <b> can be a number  or [ax+b] construct
#
#	The return value is a string with fields separated by ";" characters
#	(no whitespace)
#	gain ; relation_symbol ; additive_operator ; offset_gain ; 
#		offset_duration_symbol ; offset_additive_operator ; offset_offset
#
#	if any field is missing in the original string, then that placeholder
#	will have "null" (no quotes) in it.
#
#	example:		&
#	all placeholders will be null except for the 2nd, which will have "&"
#
#	example:		2.0@
#	place 1 will have "2.0", place 2 will have "@", all others null
#
#	example:		o+1.2
#	place 2 will have "o", place 7 (offset_offset) will have "1.2", all others
#	will be null
#
#	The calling function will then separate the fields by using 
#	split(result, ary, ";") on the result
#


	#
	#	first, eliminate ambiguity by finding any "[" or ']" which
	#	might contain a duration symbol
	#
	#	we will take care of handling fields a4,a5,a6,a7 first
	#
	i = index(rel, "[")
	if (i > 0) {
		#
		#	yes, there is a duration-bracket
		#

		#
		#	get the string between the "[" and ']"
		#
		j = index(rel, "]")
		if (j == 0) {
			print "ERROR: parse_relation_or_duration_string: incomplete duration-bracket in relation " rel | "cat 1>&2" 
			exit 1
		}
		else {
			#
			#	length of substring
			#
			len = j-i-1
			#
			#	get the sub string
			#
			srel = substr(rel, i+1, len)
			#
			#	call ourselves to parse this substring
			#
			sparse = parse_relation_or_duration_string(srel)
			#
			#	split this subparse
			#
			len = split(sparse, urel, ";")
			#
			# 	we are interested in 1,2,3, and 7
			#
			a4 = urel[1]
			a5 = urel[2]
			a6 = urel[3]
			a7 = urel[7]
		}

	}
	else {
		#
		#	no duration-bracket -- a4,a5,a6 are automatically null
		#
		a4 = "null"
		a5 = "null"
		a6 = "null"

		#
		#	now any offset is automatically a7
		#
		j = index(rel, "+")
		if (j > 0) {
			#
			#	get everything from the "+" to EOS
			#
			srel = substr(rel, j+1)
		}
		else {
			#
			#	minus?
			#
			j = index(rel, "-")
			if (j > 0) {
				#
				#	get everything from the "-" to EOS
				#
				srel = substr(rel, j+1)
			}
			else {
				srel = "null"
			}
		} # else -- minus

		a7 = srel

	} # else -- no bracket


	#
	#	now find a1,a2,a3
	#
	
	#
	#	if "i" was nonzero above, then it is the index of the "["
	#
	if (i > 0) {

		#
		#	move back from the "[" to the additive operator
		#	and this also becomes the length up the '["
		#
		i = i - 1
		srel = substr(rel, 1, i)

	} # if -- i
	else {
		#
		#	there was no bracket
		#
		srel = rel
	}

	#
	#	now srel has no ambiguity from the bracket
	#

	#	--------------
	#	test for "&"
	#	--------------
	#
	j = index(srel, "&")
	if (j) {

		#
		#	determine the gain
		#
		if (j == 1) {
			#
			#	there is no gain
			#
			a1 = "null"
		}
		else {
			#
			#	pull out the gain
			#
			a1 = substr(srel, 1, j-1)

		} # else -- has a gain

		#
		#	obvious what a2 is
		#
		a2 = "&"

		#
		#	additive operator, if any
		#
		len = length(srel)
		if (len > j) {
			a3 = substr(srel, j+1, 1)
		}
		else {
			a3 = "null"
		}


		result = sprintf("%s;%s;%s;%s;%s;%s;%s",a1,a2,a3,a4,a5,a6,a7)
		return result

	} # if -- has an "&"


	#	--------------
	#	test for "@"
	#	--------------
	#
	j = index(srel, "@")
	if (j) {

		#
		#	determine the gain
		#
		if (j == 1) {
			#
			#	there is no gain
			#
			a1 = "null"
		}
		else {
			#
			#	pull out the gain
			#
			a1 = substr(srel, 1, j-1)

		} # else -- has a gain

		#
		#	obvious what a2 is
		#
		a2 = "@"

		#
		#	additive operator, if any
		#
		len = length(srel)
		if (len > j) {
			a3 = substr(srel, j+1, 1)
		}
		else {
			a3 = "null"
		}


		result = sprintf("%s;%s;%s;%s;%s;%s;%s",a1,a2,a3,a4,a5,a6,a7)
		return result

	} # if -- has an "@"


	#	--------------
	#	test for "o"
	#	--------------
	#
	j = index(srel, "o")
	if (j) {

		#
		#	determine the gain
		#
		if (j == 1) {
			#
			#	there is no gain
			#
			a1 = "null"
		}
		else {
			#
			#	pull out the gain
			#
			a1 = substr(srel, 1, j-1)

		} # else -- has a gain

		#
		#	obvious what a2 is
		#
		a2 = "o"

		#
		#	additive operator, if any
		#
		len = length(srel)
		if (len > j) {
			a3 = substr(srel, j+1, 1)
		}
		else {
			a3 = "null"
		}


		result = sprintf("%s;%s;%s;%s;%s;%s;%s",a1,a2,a3,a4,a5,a6,a7)
		return result

	} # if -- has an "o"


	#	--------------
	#	test for "h"
	#	--------------
	#
	j = index(srel, "h")
	if (j) {

		#
		#	determine the gain
		#
		if (j == 1) {
			#
			#	there is no gain
			#
			a1 = "null"
		}
		else {
			#
			#	pull out the gain
			#
			a1 = substr(srel, 1, j-1)

		} # else -- has a gain

		#
		#	obvious what a2 is
		#
		a2 = "h"

		#
		#	additive operator, if any
		#
		len = length(srel)
		if (len > j) {
			a3 = substr(srel, j+1, 1)
		}
		else {
			a3 = "null"
		}


		result = sprintf("%s;%s;%s;%s;%s;%s;%s",a1,a2,a3,a4,a5,a6,a7)
		return result

	} # if -- has an "h"



	#	--------------
	#	test for "q"
	#	--------------
	#
	j = index(srel, "q")
	if (j) {

		#
		#	determine the gain
		#
		if (j == 1) {
			#
			#	there is no gain
			#
			a1 = "null"
		}
		else {
			#
			#	pull out the gain
			#
			a1 = substr(srel, 1, j-1)

		} # else -- has a gain

		#
		#	obvious what a2 is
		#
		a2 = "q"

		#
		#	additive operator, if any
		#
		len = length(srel)
		if (len > j) {
			a3 = substr(srel, j+1, 1)
		}
		else {
			a3 = "null"
		}


		result = sprintf("%s;%s;%s;%s;%s;%s;%s",a1,a2,a3,a4,a5,a6,a7)
		return result

	} # if -- has an "q"



	#	--------------
	#	test for "*"
	#	--------------
	#
	j = index(srel, "*")
	if (j) {

		#
		#	determine the gain
		#
		if (j == 1) {
			#
			#	there is no gain
			#
			a1 = "null"
		}
		else {
			#
			#	pull out the gain
			#
			a1 = substr(srel, 1, j-1)

		} # else -- has a gain

		#
		#	obvious what a2 is
		#
		a2 = "*"

		#
		#	additive operator, if any
		#
		len = length(srel)
		if (len > j) {
			a3 = substr(srel, j+1, 1)
		}
		else {
			a3 = "null"
		}


		result = sprintf("%s;%s;%s;%s;%s;%s;%s",a1,a2,a3,a4,a5,a6,a7)
		return result

	} # if -- has an "*"


	#	--------------
	#	test for "!"
	#	--------------
	#
	j = index(srel, "!")
	if (j) {

		#
		#	determine the gain
		#
		if (j == 1) {
			#
			#	there is no gain
			#
			a1 = "null"
		}
		else {
			#
			#	pull out the gain
			#
			a1 = substr(srel, 1, j-1)

		} # else -- has a gain

		#
		#	obvious what a2 is
		#
		a2 = "!"

		#
		#	additive operator, if any
		#
		len = length(srel)
		if (len > j) {
			a3 = substr(srel, j+1, 1)
		}
		else {
			a3 = "null"
		}


		result = sprintf("%s;%s;%s;%s;%s;%s;%s",a1,a2,a3,a4,a5,a6,a7)
		return result

	} # if -- has an "!"

	return "null"
}
#--------------------- parse_relation_or_duration_string ---------------



function collapse_relation_parse_to_string(a1, a2, a3, a4, a5, a6, a7,   result)
{
#
#	the reverse of parse_relation_or_duration_string()
#
#	this output string is of the form seen by the user
#

	if (debug_child) {
		print "DEBUGGER: collapse_relation_parse_to_string: input: a1= (" a1 "), a2= (" a2 "), a3= (" a3 "), a4= (" a4 "), a5= (" a5 "), a6= (" a6 "), a7= (" a7 ")"
	}

	#
	#	gain and first symbol
	#
	if (a1 == "null" || a1 == 1) {

		result = sprintf("%s", a2);

	}
	else {

		result = sprintf("%s%s", a1, a2);

	}


	#
	#	is there an offset?
	#
	if (a3 != "null") {

		#
		#	put in the additive operator
		#
		if ((a3 == "+") || (a3 == "-")) {
			result = sprintf("%s%s", result, a3);
		}
		else {
			print "ERROR: collapse_relation_parse_to_string: unsupported additive operator (" a3 ") in (" a1 a2 a3 a4 a5 a6 a7 ")" | "cat 1>&2" 
			exit 1
		}

		#
		# should there be brackets?
		#
		if (a5 == "null") {
			#
			#	there is no duration operator in the offset
			#
			if (a7 == "null") {
				print "ERROR: collapse_relation_parse_to_string: additive operator (" a3 ") but no offset" | "cat 1>&2" 
				exit 1
			}
			result = sprintf("%s%s", result, a7);
		}
		else {
			#
			#	a bracket is needed for the offset
			#
			result = sprintf("%s[", result)

			#
			#	the gain within the brackets
			#
			if (a4 == "null" || a4 == 1) {
				result = sprintf("%s%s", result, a5)
			}
			else {
				result = sprintf("%s%s%s", result, a4,a5)
			}

			#
			#	an offset within the brackets?
			#
			if (a6 != "null" && a7 != 0) {

				if (a6 == "+" || a6 == "-") {
					result = sprintf("%s%s%s]", result, a6,a7)
				}
				else {
					print "ERROR: collapse_relation_parse_to_string: unsupported additive operator (" a6 ")" | "cat 1>&2" 
					exit 1
				}
			}


		} # else -- a bracket is needed

	}
	return result
}
#------------------ collapse_relation_parse_to_string ------------------



function multiply_relation_symbol(insymbol, multiplier,   len,a,sparse,a1,a2,a3,a4,a5,a6,a7,result)
{
#
#	Called by Multiply_ScVar
#
#	we expect that "multiplier" has been resolved into a literal by the
#	calling function
#
#	note:  if the original relation symbol had a construction with
#	[q], [1.2h+1.5], ... kind of offset, this symbolic level is lost
#	here and turned into a literal offset number
#

	sparse = parse_relation_or_duration_string(insymbol)
	if (sparse == "null") {
		print "ERROR: multiply_relation_symbol: (" insymbol ") not a relation" | "cat 1>&2" 
		exit 1
	}

	if (debug_child) {
	print "DEBUGGER: multiply_relation_symbol: parse'd string= (" sparse ")"
	}

	len = split(sparse, a, ";")
	if (len != 7) {
		print "ERROR: multiply_relation_symbol: not 7 parse fields" | "cat 1>&2" 
		exit 1
	}

	if (debug_child) {
	print "DEBUGGER: multiply_relation_symbol: split into (" len ") substrings"
	print "DEBUGGER: multiply_relation_symbol: a[1]= (" a[1] ")"
	print "DEBUGGER: multiply_relation_symbol: a[2]= (" a[2] ")"
	print "DEBUGGER: multiply_relation_symbol: a[3]= (" a[3] ")"
	print "DEBUGGER: multiply_relation_symbol: a[4]= (" a[4] ")"
	print "DEBUGGER: multiply_relation_symbol: a[5]= (" a[5] ")"
	print "DEBUGGER: multiply_relation_symbol: a[6]= (" a[6] ")"
	print "DEBUGGER: multiply_relation_symbol: a[7]= (" a[7] ")"
	}

	#
	#	a[1]	=	gain
	#	a[2]	=	symbol
	#	a[3]	=	additive operator
	#	a[4]	= 	offset_gain
	#	a[5]	=	offset_symbol
	#	a[6]	=	offset_additive_operator
	#	a[7]	=	offset_offset
	#

	#
	#	defaults:
	#
	a1 = "null"
	a2= "null"
	a3 = "null"
	a4 = "null"
	a5 = "null"
	a6 = "null"
	a7 = "null"

	#
	#	multiply numbers, then collapse back into a string
	#

	if (a[1] == "null") {
		a1 = multiplier
	}
	else {
		a1 = a[1] * multiplier
	}

	if (debug_child) {
	print "DEBUGGER: multiply_relation_symbol: a1= (" a1 ")"
	}


	a2 = a[2]
	a3 = a[3]

	if (a[5] != "null") {
		#
		#	there was a duration symbol in the offset
		#
		if (a[4] == "null") {
			a4 = multiplier
		}
		else {
			a4 = a[4] * multiplier
		}

		a5 = a[5]

		if (a[6] != "null") {
			if (a[7] != "null") {
				a7 = a[7] * multiplier
				a6 = a[6]
			}
		}

	}

	if (a[7] != "null") {
		a7 = a[7] * multiplier
	}


	result = collapse_relation_parse_to_string(a1,a2,a3,a4,a5,a6,a7)
	return result
}
#--------------------- multiply_relation_symbol ---------------------------



function add_to_relation_symbol(insymbol, multiplier,   len,a,sparse,a1,a2,a3,a4,a5,a6,a7,result)
{
#
#	Called by Add_ScVar
#
#	we expect that "multiplier" has been resolved into a literal by the
#	calling function
#
#	note:  if the original relation symbol had a construction with
#	[q], [1.2h+1.5], ... kind of offset, this symbolic level is lost
#	here and turned into a literal offset number
#

	sparse = parse_relation_or_duration_string(insymbol)
	if (sparse == "null") {
		print "ERROR: add_to_relation_symbol: (" insymbol ") not a relation" | "cat 1>&2" 
		exit 1
	}

	if (debug_child) {
	print "DEBUGGER: add_to_relation_symbol: parse'd string= (" sparse ")"
	}

	len = split(sparse, a, ";")
	if (len != 7) {
		print "ERROR: add_to_relation_symbol: not 7 parse fields" | "cat 1>&2" 
		exit 1
	}

	if (debug_child) {
	print "DEBUGGER: add_to_relation_symbol: split into (" len ") substrings"
	print "DEBUGGER: add_to_relation_symbol: a[1]= (" a[1] ")"
	print "DEBUGGER: add_to_relation_symbol: a[2]= (" a[2] ")"
	print "DEBUGGER: add_to_relation_symbol: a[3]= (" a[3] ")"
	print "DEBUGGER: add_to_relation_symbol: a[4]= (" a[4] ")"
	print "DEBUGGER: add_to_relation_symbol: a[5]= (" a[5] ")"
	print "DEBUGGER: add_to_relation_symbol: a[6]= (" a[6] ")"
	print "DEBUGGER: add_to_relation_symbol: a[7]= (" a[7] ")"
	}

	#
	#	a[1]	=	gain
	#	a[2]	=	symbol
	#	a[3]	=	additive operator
	#	a[4]	= 	offset_gain
	#	a[5]	=	offset_symbol
	#	a[6]	=	offset_additive_operator
	#	a[7]	=	offset_offset
	#

	#
	#	defaults:
	#
	a1 = a[1]
	a2 = a[2]
	a4 = a[4]
	a5 = a[5]


	if (a[5] != "null") {

		#
		#	if there are brackets, then we don't worry about
		#	a1,a2,a3
		#
		a3 = a[3]

		if (a[7] == "null") {
			a7 = multiplier
			a6 = "+"
		}
		else if (a[6] == "+") {
			a7 = a[7] + multiplier
			a6 = a[6]
		}
		else if (a[6] == "-") {
			a7 = multiplier - a[7]
			a6 = a[6]
		}


	}
	else {

		#
		#	if there are no brackets, then we don't worry about
		#	a4,a5,a6
		#
		a6 = a[6]

		if (a[7] == "null") {
			a7 = multiplier
			if (a[2] == "null") {
				a3 = "null"
			}
			else {
				a3 = "+"
			}
		}
		else if (a[3] == "+") {
			a7 = a[7] + multiplier
			a3 = a[3]
		}
		else if (a[3] == "-") {
			a7 = multiplier - a[7]
			a3 = a[3]
		}


	}



	result = collapse_relation_parse_to_string(a1,a2,a3,a4,a5,a6,a7)
	return result
}
#--------------------- add_to_relation_symbol ---------------------------



function get_string_matrix_rowscols(sm,   i,ary,cary,rows,cols,cl)
{

	rows = split(sm, ary, ";")
	cols = 1
	for (i in ary) {
		cl = split(ary[i], cary, " ")
		if (cl > cols) {
			cols = cl
		}
	}
	return sprintf("%d %d", rows,cols)
}
#----------------- get_string_matrix_rowscols ----------------------------



function get_element_from_string_matrix(sm, row, col,   as,am,bs,bm)
{

	#
	#	split into rows
	#
	as = split(sm, am, ";")
	if (row <= as) {

		bs = split(am[row], bm, " ")
		if (col <= bs) {
			return bm[col]
		}
	}

	return "null"
}
#---------------- get_element_from_string_matrix -------------------------



function transpose_string_matrix(sm,   i,j,el,mrc,maxrc,result)
{
	#
	#	get the dimensions
	#
	mrc = get_string_matrix_rowscols(sm)
	split(mrc,maxrc," ")

	#
	#	build the new matrix row-wise by copying 
	#	(r,c) from (c,r), (r,c+1) from (c+1,r), etc
	#
	for (i=1; i<=maxrc[2]; i++) {

		#
		#	loop down rows of a column
		#
		for (j=1; j<=maxrc[1]; j++) {

			el = get_element_from_string_matrix(sm, j, i)
			result = sprintf("%s %s", result, el)

		}

		if (i < maxrc[2]) {
			result = sprintf("%s ;", result)
		}

	} # for -- i

	return result
}
#-------------------- transpose_string_matrix ---------------------------



function print_string_matrix(sm,   i,as,ary)
{
	as = split(sm, ary, ";")

	for (i=1; i<=as; i++) {

		print ary[i]

	}
	return
}
#--------------------- print_string_matrix ------------------------------



function increment_name(oldname,   i,j,k,len,cvar,nvar,newname)
{

	cvar = oldname
	len = length(cvar)
	i = match(cvar, /(_m[0-9])+/)
	if (i) {

		#
		#	skip to the last "_"
		#
		k = i+2
		nvar = substr(cvar, k)
		j = match(nvar, /_m[0-9]+/)
		while (j) {

			k = j+2
			nvar = substr(nvar, k)
			j = match(nvar, /_m[0-9]+/)

		}

		#
		#	k should be the index of the last number
		#	nvar should be the number itself
		#
		cvar = substr(cvar, 1, k-3)		# stem
		nvar += 1
		newname = sprintf("%s_%d", cvar, nvar)


	}
	else {
		newname = sprintf("%s_m1", cvar)
	}

	#
	#	check if name already exists
	#
	k = get_var_index_from_name(newname)
	while (k > 0) {

		i = match(newname, /_m[0-9]+/)
		j = i+2
		nvar = substr(newname, j)
		nvar += 1
		cvar = substr(newname, 1, j-3)		# stem
		newname = sprintf("%s_m%d", cvar, nvar)
		k = get_var_index_from_name(newname)
	}


	return newname
}
#-------------------- increment_name ------------------------------------



function add_halftones_to_key(adder, oldkey,    mant,dec,oldtones,newtones,newmant,newdec,result)
{

	#
	#	decompose oldkey into halftones
	#
	mant = int(oldkey)
	dec = oldkey - mant
	dec = dec * 100



	oldtones = mant * 12
	oldtones += dec

	if (debug_child) {
		print "DEBUGGER:  old mant = " mant ", old dec = " dec ", oldtones = " oldtones
	}

	#
	#	adder is assumed to already be expressed in halftones
	#
	newtones = oldtones + adder


	newmant = int(newtones / 12)
	newdec = newtones - (newmant*12)

	if (debug_child) {
		print "DEBUGGER:  newtones = " newtones ", new mant = " newmant ", new dec = " newdec
	}

	result = newmant + (newdec/100)

	if (debug_child) {
		print "DEBUGGER: new key = " result
	}

	return result
}
#------------------ add_halftones_to_key----------------------------------



function Multiply_ScVar(oldcolumn, multiplier,    i,j,k,cvar,nvar,tvar,as,ary,nstr,result,mt,mz,var)
{
#
#	Called by Mod_ScVar.
#	Given the column of values to be modified, "oldcolumn", multiplies
#	the values by "multiplier" and returns a new column as a variable name.
#

	#
	#	at first, "oldcolumn" is just a variable name with a "^" on it
	#
	if (match(oldcolumn, /\^/)) {

		#
		# take off the leading "^"
		#
		cvar = substr(oldcolumn,2)

	} # if -- match
	else {
		cvar = oldcolumn
	}

	if (debug_child) {
	print "DEBUGGER: Multiply_ScVar: old column = (" cvar ")"
	}

	#
	#	create the new name (myvar --> myvar_1,  myvar_1 --> myvar_2, etc)
	#
	nvar = increment_name(cvar)

	if (debug_child) {
	print "DEBUGGER: Multiply_ScVar: new column name = (" nvar ")"
	}

	#
	#	resolve the multiplier if it is a variable
	#
	if (is_a_numeric_var(multiplier)) {
		mt = multiplier
	}
	else {
		k = get_var_index_from_name(multiplier)
		mt = var_value[k]
	}
	#	---------------------------------------
	#	now, the multiplier is a literal in "mt"
	#	---------------------------------------

	if (debug_child) {
	print "DEBUGGER: multiplier= (" mt ")"	
	}

	#
	#	spilt the column into its fields
	#
	k = get_var_index_from_name(cvar)
	as = split(var_value[k], ary, " ")

	if (debug_child) {
	#	DEBUGGER
	print "DEBUGGER: column has components: "
	for (i=1; i<=as; i++) {
		print "DEBUGGER: " ary[i]
	}
	}


	#
	#	multiply each field
	#
	for (i=1; i<=as; i++) {

		#
		#	resolve the field if it is a variable
		#
		if (is_a_numeric_var(ary[i])) {
			#
			#	do the multiplication
			#
			mz = mt * ary[i]

			if (debug_child) {
			print "DEBUGGER: multiplied numeric = (" mz ")"
			}
		}
		else if (is_a_relation_symbol(ary[i])) {
			#
			#	must handle this separately
			#	multiply the gain
			#	multiply the offset
			#	return a new relation symbol
			#
			mz = multiply_relation_symbol(ary[i], mt)

			if (debug_child) {
			print "DEBUGGER: multiplied relation = (" mz ")"
			}
		}
		else if (is_a_var_reference(ary[i])) {
			#
			#	find the variable after the "$" prefix
			#
			j = get_var_index_from_reference(ary[i])
			#
			#	do the multiplication with the referenced value
			#
			mz = mt * var_value[j]

			if (debug_child) {
			print "DEBUGGER: multiplied reference = (" mz ")"
			}

		}
		else if (is_an_operation(ary[i])) {
			#
			# if this was created by "eval"
			#
			#	we must create a new variable to hold this
			#	because it involve a set of references
			#	that must be accessed at matrix resolution time
			#
			#
			#	new name
			#
			tvar = increment_name(ary[i])

			#
			#	find this variable in the table
			#
			k = get_var_index_from_name(ary[i])


			#
			# new variable for this new "operation" variable
			#
			varcount += 1				# one more var in table
			j = varcount
			var_name[j] = tvar
			var_value[j] = var_value[k]
			var_size[j] = 1
			var_op[j] = var_op[k]
			var_operand[j] = var_operand[k] * mt
			var_reference[j] = var_reference[k]

			mz = tvar

			if (debug_child) {
			print "DEBUGGER: multiplied operation = (" mz ")"
			}

		}
		else {
			#
			# unsupported
			#
			print "ERROR: Multiply_ScVar: unsupported column element (" ary[i] ")" | "cat 1>&2" 
			exit 1
		}


		#
		#	append this multiplied value to the list of values
		#
		nstr = sprintf("%s %s", nstr, mz)

	} # for -- i

	#
	# new variable for this new column
	#
	varcount += 1				# one more var in table
	j = varcount
	var_name[j] = nvar
	var_value[j] = nstr
	var_size[j] = as
	var_op[j] = "null"
	var_operand[j] = "null"
	var_reference[j] = "null"

	#
	#	attach the column operator to match the input variable
	#
	result = sprintf("^%s", var_name[j])

	if (debug_child) {
	print "DEBUGGER: Multiply_ScVar: returning new var (" result ")"
	}

	#
	#	done
	#
	return result
}
#------------------------ Multiply_ScVar --------------------------------



function Add_ScVar(oldcolumn, multiplier,    i,j,k,cvar,nvar,tvar,as,ary,nstr,result,mt,mz,var)
{
#
#	Called by Mod_ScVar.
#	Given the column of values to be modified, "oldcolumn", adds
#	the values by "multiplier" and returns a new column as a variable name.
#

	#
	#	at first, "oldcolumn" is just a variable name with a "^" on it
	#
	if (match(oldcolumn, /\^/)) {

		#
		# take off the leading "^"
		#
		cvar = substr(oldcolumn,2)

	} # if -- match
	else {
		cvar = oldcolumn
	}

	if (debug_child) {
	print "DEBUGGER: Add_ScVar: old column = (" cvar ")"
	}

	#
	#	create the new name (myvar --> myvar_1,  myvar_1 --> myvar_2, etc)
	#
	nvar = increment_name(cvar)

	if (debug_child) {
	print "DEBUGGER: Add_ScVar: new column name = (" nvar ")"
	}

	#
	#	resolve the multiplier if it is a variable
	#
	if (is_a_numeric_var(multiplier)) {
		mt = multiplier
	}
	else {
		k = get_var_index_from_name(multiplier)
		mt = var_value[k]
	}
	#	---------------------------------------
	#	now, the multiplier is a literal in "mt"
	#	---------------------------------------

	if (debug_child) {
	print "DEBUGGER: multiplier= (" mt ")"	
	}

	#
	#	spilt the column into its fields
	#
	k = get_var_index_from_name(cvar)
	as = split(var_value[k], ary, " ")

	if (debug_child) {
	#	DEBUGGER
	print "DEBUGGER: column has components: "
	for (i=1; i<=as; i++) {
		print "DEBUGGER: " ary[i]
	}
	}


	#
	#	add to each field
	#
	for (i=1; i<=as; i++) {

		#
		#	resolve the field if it is a variable
		#
		if (is_a_numeric_var(ary[i])) {
			#
			#	do the multiplication
			#
			mz = mt + ary[i]

			if (debug_child) {
			print "DEBUGGER: added to numeric = (" mz ")"
			}
		}
		else if (is_a_relation_symbol(ary[i])) {
			#
			#	must handle this separately
			#	multiply the gain
			#	multiply the offset
			#	return a new relation symbol
			#
			mz = add_to_relation_symbol(ary[i], mt)

			if (debug_child) {
			print "DEBUGGER: added to relation = (" mz ")"
			}
		}
		else if (is_a_var_reference(ary[i])) {
			#
			#	find the variable after the "$" prefix
			#
			j = get_var_index_from_reference(ary[i])
			#
			#	do the multiplication with the referenced value
			#
			mz = mt + var_value[j]

			if (debug_child) {
			print "DEBUGGER: added to reference = (" mz ")"
			}

		}
		else if (is_an_operation(ary[i])) {
			#
			# if this was created by "eval"
			#
			#	we must create a new variable to hold this
			#	because it involve a set of references
			#	that must be accessed at matrix resolution time
			#
			#
			#	new name
			#
			tvar = increment_name(ary[i])

			#
			#	find this variable in the table
			#
			k = get_var_index_from_name(ary[i])


			#
			# new variable for this new "operation" variable
			#
			varcount += 1				# one more var in table
			j = varcount
			var_name[j] = tvar
			var_value[j] = var_value[k]
			var_size[j] = 1
			var_op[j] = var_op[k]
			var_operand[j] = var_operand[k] + mt
			var_reference[j] = var_reference[k]

			mz = tvar

			if (debug_child) {
			print "DEBUGGER: added to operation = (" mz ")"
			}

		}
		else {
			#
			# unsupported
			#
			print "ERROR: Add_ScVar: unsupported column element (" ary[i] ")" | "cat 1>&2" 
			exit 1
		}


		#
		#	append this multiplied value to the list of values
		#
		nstr = sprintf("%s %s", nstr, mz)

	} # for -- i

	#
	# new variable for this new column
	#
	varcount += 1				# one more var in table
	j = varcount
	var_name[j] = nvar
	var_value[j] = nstr
	var_size[j] = as
	var_op[j] = "null"
	var_operand[j] = "null"
	var_reference[j] = "null"

	#
	#	attach the column operator to match the input variable
	#
	result = sprintf("^%s", var_name[j])

	if (debug_child) {
	print "DEBUGGER: Add_ScVar: returning new var (" result ")"
	}

	#
	#	done
	#
	return result
}
#------------------------ Add_ScVar --------------------------------



function KeyShift_ScVar(oldcolumn, multiplier,    i,j,k,cvar,nvar,tvar,as,ary,nstr,result,mt,mz,var)
{
#
#	Called by Mod_ScVar.
#	Given the column of values to be modified, "oldcolumn", adds
#	the values by "multiplier" and returns a new column as a variable name.
#

	#
	#	at first, "oldcolumn" is just a variable name with a "^" on it
	#
	if (match(oldcolumn, /\^/)) {

		#
		# take off the leading "^"
		#
		cvar = substr(oldcolumn,2)

	} # if -- match
	else {
		cvar = oldcolumn
	}

	if (debug_child) {
	print "DEBUGGER: KeyShift_ScVar: old column = (" cvar ")"
	}

	#
	#	create the new name (myvar --> myvar_1,  myvar_1 --> myvar_2, etc)
	#
	nvar = increment_name(cvar)

	if (debug_child) {
	print "DEBUGGER: KeyShift_ScVar: new column name = (" nvar ")"
	}

	#
	#	resolve the multiplier if it is a variable
	#
	if (is_a_numeric_var(multiplier)) {
		mt = multiplier
	}
	else {
		k = get_var_index_from_name(multiplier)
		mt = var_value[k]
	}
	#	---------------------------------------
	#	now, the multiplier is a literal in "mt"
	#	---------------------------------------

	if (debug_child) {
	print "DEBUGGER: multiplier= (" mt ")"	
	}

	#
	#	spilt the column into its fields
	#
	k = get_var_index_from_name(cvar)
	as = split(var_value[k], ary, " ")

	if (debug_child) {
	#	DEBUGGER
	print "DEBUGGER: column has components: "
	for (i=1; i<=as; i++) {
		print "DEBUGGER: " ary[i]
	}
	}


	#
	#	add to each field
	#
	for (i=1; i<=as; i++) {

		#
		#	resolve the field if it is a variable
		#
		if (is_a_numeric_var(ary[i])) {
			#
			#	do the multiplication
			#
			mz = add_halftones_to_key(mt, ary[i])

			if (debug_child) {
			print "DEBUGGER: added to numeric = (" mz ")"
			}
		}
		else if (is_a_relation_symbol(ary[i])) {
			#
			#	must handle this separately
			#	multiply the gain
			#	multiply the offset
			#	return a new relation symbol
			#
			print "ERROR: Shift Key in child function doesn't take relation expressions (" ary[i] ")" | "cat 1>&2" 
			exit 1
		}
		else if (is_a_var_reference(ary[i])) {
			#
			#	find the variable after the "$" prefix
			#
			j = get_var_index_from_reference(ary[i])
			#
			#	do the multiplication with the referenced value
			#
			mz = add_halftones_to_key(mt, var_value[j])

			if (debug_child) {
			print "DEBUGGER: added to reference = (" mz ")"
			}

		}
		else if (is_an_operation(ary[i])) {
			#
			# if this was created by "eval"
			#
			#	we must create a new variable to hold this
			#	because it involve a set of references
			#	that must be accessed at matrix resolution time
			#
			#
			#	new name
			#
			tvar = increment_name(ary[i])

			#
			#	find this variable in the table
			#
			k = get_var_index_from_name(ary[i])


			#
			# new variable for this new "operation" variable
			#
			varcount += 1				# one more var in table
			j = varcount
			var_name[j] = tvar
			var_value[j] = var_value[k]
			var_size[j] = 1
			var_op[j] = var_op[k]
			var_operand[j] = add_halftones_to_key(mt, var_operand[k])
			var_reference[j] = var_reference[k]

			mz = tvar

			if (debug_child) {
			print "DEBUGGER: added to operation = (" mz ")"
			}

		}
		else {
			#
			# unsupported
			#
			print "ERROR: KeyShift_ScVar: unsupported column element (" ary[i] ")" | "cat 1>&2" 
			exit 1
		}


		#
		#	append this multiplied value to the list of values
		#
		nstr = sprintf("%s %s", nstr, mz)

	} # for -- i

	#
	# new variable for this new column
	#
	varcount += 1				# one more var in table
	j = varcount
	var_name[j] = nvar
	var_value[j] = nstr
	var_size[j] = as
	var_op[j] = "null"
	var_operand[j] = "null"
	var_reference[j] = "null"

	#
	#	attach the column operator to match the input variable
	#
	result = sprintf("^%s", var_name[j])

	if (debug_child) {
	print "DEBUGGER: KeyShift_ScVar: returning new var (" result ")"
	}

	#
	#	done
	#
	return result
}
#------------------------ KeyShift_ScVar --------------------------------



function Mod_ScVar(newname, oldname, argstring,   i,ind,nind,as,ary,argary,coln,op,newcol)
{
#
#	This is what modifies a pattern variable, by the script command
#
#		child  newname  oldname  column_to_mod  operator  arg1  arg2 ....
#
#	which only works on a variable that is ready for the score line, i.e.,
#	represents a csound matrix. Thus, we take an existing note pattern
#	and change one of its csound parameters (one column).  For example,
#	we might transpose its key by operating on that particular column.
#	We might change its tempo by operating on the first column, or by
#	operating on the fourth (csound's third) column.
#
#	"argstring" is taken from the script is everything from "column_to_mod"
#	to the right
#
#	Doesn't have to return anything: we already know the new name.  It will
#	now be listed in the symbol table
#

	if (debug_child) {
	print "DEBUGGER: Mod_ScVar: newname= (" newname "), oldname= (" oldname "), argstring= (" argstring ")"
	}

	#
	#	"oldname" is the name of the old pattern variable to be child'ed
	#	find it in the Variable list
	#
	i = get_var_index_from_name(oldname)
	if (var_size[i] <= 1) {
		print "ERROR: Mod_ScVar: variable " oldname " is not a pattern" | "cat 1>&2" 
		exit 1
	}
	ind = i


	#
	#	get the arguments -- a partial length of the original argument string
	#	is given as "argstring"
	#
	as = split(argstring, argary, " ")
	coln = argary[1] 	# column number should be first
	op = argary[2]		#	operator (+,-,*,/,...other)

	if (debug_child) {
	print "DEBUGGER: Mod_ScVar: column= (" coln "), op= (" op"), multiplier= (" argary[3] ")"
	}


	#
	#	the "value" of the old pattern name should be a row of sub-variables
	#	each sub-variable should be a column (should have a ^ character prefixed)
	#	split into "columns"
	#
	as = split(var_value[ind], ary, " ")
	if (as != var_size[ind]) {
		print "ERROR: Mod_ScVar: num fields found for variable " sfield " not equal " var_size[ind] | "cat 1>&2" 
		exit 1
	}

	if (debug_child) {
	#	DEBUGGER
	print "DEBUGGER: old pattern= ("
	for (i=1; i<=as; i++) {
		print ary[i]
	}
	print ")"
	}


	#
	#	create the new variable to hold the new pattern
	#
	#
	#
	#	check if there is already a variable with that name
	#
	for (i=1; i<varcount; i++) {
		if (newname == var_name[i]) {
			print "ERROR: Mod_ScVar: new name " newname " is already assigned" | "cat 1>&2" 
			exit 1
		}
	}

	#
	# new variable, increment count
	#
	varcount += 1
	nind = varcount
	var_name[nind] = newname
	var_op[nind] = "null"
	var_operand[nind] = "null"
	var_reference[nind] = "null"

	#
	#	CAUTION: don't use varcount (or varindex) again to access this variable
	#	because it will be changed in the subfunctions called below
	#


	#
	#	Dup all the columns except the one to be modified
	#
	for (i=1; i<=as; i++) {

		if (i == coln) {

			if (op == "*") {
				if (debug_child) {
				print "DEBUGGER: Mod_ScVar: mulyiply to column (" ary[i] ")"
				}

				newcol = Multiply_ScVar(ary[i], argary[3])

				if (debug_child) {
				print "DEBUGGER: Mod_ScVar: received new multiplied column (" newcol ")"
				}

			}
			else if (op == "+") {
				if (debug_child) {
				print "DEBUGGER: Mod_ScVar: add to column (" ary[i] ")"
				}

				newcol = Add_ScVar(ary[i], argary[3])

				if (debug_child) {
				print "DEBUGGER: Mod_ScVar: received new added-to column (" newcol ")"
				}

			}
			else if (op == "k+") {
				if (debug_child) {
				print "DEBUGGER: Mod_ScVar: key shift column (" ary[i] ")"
				}

				newcol = KeyShift_ScVar(ary[i], argary[3])

				if (debug_child) {
				print "DEBUGGER: Mod_ScVar: received new key shift column (" newcol ")"
				}

			}
			else {
				print "ERROR: unknown op (" op ")" | "cat 1>&2" 
				exit 1
			}

			#
			#	the new column name returned is expected to already
			#	have an ^ prefix
			#
			var_value[nind] = sprintf("%s %s", var_value[nind], newcol)
		}
		else {
			#
			#	Dup it
			#
			var_value[nind] = sprintf("%s %s", var_value[nind], ary[i])
		}

		#
		# keep track of the size
		#
		var_size[nind] += 1

		if (debug_child) {
		print "DEBUGGER: Mod_ScVar: new var (" var_name[nind] ") has (" var_value[nind] ")"
		}

	} # for -- loop thru columns

}
#------------------------- Mod_ScVar ------------------------------------



function Eval_Variable(field,  i,j,k,sfield,ref,op,opd,refv,opdv,ret) {

#
#	On the first call...
#	"field" is the field name exactly as it is found
#	in the file.  No "$" is expected.
#
#	Recursive calls might expect a "$" or a number.
#
#	in the command "seteval name arg1 arg2 arg3":
#	arg1 can be a reference to a variable, e.g. $b1, or a number
#	arg2 must be an operation, e.g., +, -, *, /
#	arg3 can be a reference to a variable, e.g. $b1, or a number
#


	#
	#	remove any leading "$" or "^"
	#
	if (match(field, /\$/)) {
		sfield = substr(field,2)
	}
	else if (match(field, /\^/)) {
		sfield = substr(field,2)
	}
	else {
		sfield = field
	}



	if (is_in_varlist(sfield)) {

		i = get_var_index_from_name(sfield)
		ref = var_reference[i]
		op = var_op[i]
		opd = var_operand[i]

		#
		#	if the reference is itself an eval, then we must
		#	recursively call Eval_Variable on it
		#
		if (is_a_var_reference(ref)) {
			j = get_var_index_from_reference(ref)

			if (var_op[j] != "null") {
				#
				#	the reference itself is a seteval
				#
				refv = Eval_Variable(var_name[j])
			}
			else {
				#
				# then get the value
				#
				refv = var_value[j]
				
			}


		}
		else {
			#
			#	else, ref must be a number
			#
			refv = ref
		}



		#
		#	if the operand is itself an eval, then we must
		#	recursively call Eval_Variable on it
		#
		if (is_a_var_reference(opd)) {
			k = get_var_index_from_reference(opd)

			if (var_op[k] != "null") {
				#
				#	the operand itself is a seteval
				#
				opdv = Eval_Variable(var_name[k])
			}
			else {
				#
				# then get the value
				#
				opdv = var_value[k]
			}

		}
		else {
			#
			#	else, opd must be a number
			#
			opdv = opd
		}

		#
		#
		#	at this point both the ref and opd should have been
		#	resolved into numbers
		#
		if (refv == "null" ) {
			print "ERROR: Eval_Variable: refv resolved to null" | "cat 1>&2" 
			exit 1
		}
		if (opdv == "null" ) {
			print "ERROR: Eval_Variable: opdv resolved to null" | "cat 1>&2" 
			exit 1
		}


		#
		#	do the operation
		#
		if (op == "+") {
			ret = refv + opdv
		}
		else if (op == "-") {
			ret = refv - opdv
		}
		else if (op == "*") {
			ret = refv * opdv
		}
		else if (op == "/") {
			ret = refv / opdv
		}

		return ret


	}
	else {
		#
		# must be a number
		#
		return sfield
	}

}
#----------------------- Eval_Variable ---------------------------------



function is_an_alias(field,   ind,par1)
#
#	determine if the field in a "set" definition is an alias:
#	that is, is of the form  "~name(value)"
#
{

	ind = index(field, "~")
	if (ind != 1) {
		return 0	# not an alias call
	}

	#
	#	if we drop down to here, then the first character in the field
	#	was an "~", so this must be an alias call
	#
	ind = index(field, "(")
	if (ind < 3) {
		#
		#	there must be at least 1 character following the "~"
		#	that is, there must be at least one character in the alias
		#	name
		#
		print "ERROR: left paranthesis in alias call at field (" field ")" | "cat 1>&2" 
		exit 1
	}

	par1 = ind	# index of the opening paranthesis

	ind = index(field, ")")
	if (ind < (par1+2)) {
		#
		#	there must be at least 1 character following the "("
		#	that is, there must be at least one character inside
		#	the paranthesis--a void argument is not allowed
		#
		print "ERROR: right paranthesis in alias call at field (" field ")" | "cat 1>&2" 
		exit 1
	}
	
	return 1	# success
}
#	-------------------- is_an_alias -----------------------------------



function insert_alias(whole_alias,   ind,par1,par2,name,arginput)
#
#	Given an entire alias call, "~name(arg_input)", this function returns
#	the output value associated with arg_input
#
{

	if (is_an_alias(whole_alias) == 0) {
		print "ERROR (bug): insert_alias: argument is not an alias call" | "cat 1>&2" 
		exit 1
	}


	#
	#	go directly to the parantheses
	#
	ind = index(whole_alias, "(")
	par1 = ind
	ind = index(whole_alias, ")")
	par2 = ind
	name = substr(whole_alias, 2, (par1-2))
	arginput = substr(whole_alias, (par1+1), (par2-par1-1))

	if ((name, arginput) in alias_list) {
		return alias_list[name, arginput]
	}
	else {
		print "ERROR: argument (" arginput ") not found in alias (" name ")" | "cat 1>&2" 
		exit 1
	}
}
#	-------------------- insert_alias -----------------------------------



function match_skiptrack(track,   i)
{
	for (i=1; i<=tracks; i++) {
		if ((i) in skip_track) {
			if (track == skip_track[i]) {
				return 1
			}
		}
	}
	return 0
}
#	------------------- match_skiptrack ----------------------------------



function Sc_EvalVar(field,   i,iscol,sfield,islist,lresult,result) {
#
#	Eval Var in a score line
#	The return value must be a number or a matrix of numbers
#	A matrix of numbers is an array of strings
#
#	Assumes that "field" is a variable:  literal numbers are not passed
#	to this function
#
#	Since you can't return an array, the matrix returned will be
#	of the form, "a1 a2 a3... ; b1 b2 b3 ... ; ..."
#	----> "string-matrix" format
#

	#print "Sc_EvalVar receives (" field ")"

	#
	#	before we can find the name in the variable list
	#	we must check if there is a leading "^" that specifies
	#	COLUMN evaluation
	#
	if (match(field, /\^/)) {
		iscol = 1
	} # if -- match

	if (iscol) {
		#
		# take off the leading "^"
		#
		sfield = substr(field,2)
	}
	else {
		sfield = field
	}

	#print "Sc_EvalVar: sfield= " sfield

	#
	#	find it in the Variable list
	#
	for (i=1; i<=varcount; i++) {

		if (var_name[i] == sfield) {

			if (var_size[i] > 1) {
				islist = 1
			}
			#print "DEBUGGER: Sc_EvalVar: sfield value=" var_value[i]

			if (islist) {

				if (iscol) {
					#print "Sc_EvalVar passes (" sfield ") to Sc_EvalCol"
					result = Sc_EvalCol(sfield)
					#print "DEBUGGER: Sc_EvalVar: ScEvalCol returns:"
					#print_string_matrix(result)
					return result
				}
				else {
					result = Sc_EvalRow(sfield)
					#print "DEBUGGER: Sc_EvalVar: ScEvalRow returns:"
					#print_string_matrix(result)
					return result
				}

			}
			else {
				#
				#	a single field
				#
				if (var_op[i] != "null") {
					#
					#	if this is a "function" variable defined with seteval
					#	we must return the ultimate value resolved by the
					#	nested operations (run Eval_Variable)
					#
					#print "DEBUGGER: seteval var " var_name[i]
					return Eval_Variable(var_name[i])
					
				}
				else if (is_a_numeric_var(var_value[i])) {
					#print "DEBUGGER: Sc_EvalVar: numeric= " var_value[i]
					return var_value[i]
				}
				else if (is_a_relation_symbol(var_value[i])) {
					#print "DEBUGGER: Sc_EvalVar: relation= " var_value[i]
					return var_value[i]
				}
				else {
					#
					#	recuresively call Sc_EvalVar until
					#	a number, or array of strings, is found
					#
					result = Sc_EvalVar(var_value[i])
					lresult = result
					while (is_a_numeric_var(result)==0 && is_a_relation_symbol(result)==0) {
						result = Sc_EvalVar(result)
						if (result == "null") {
							return lresult
						}
					}
					return result
				}

			} # else -- a single field

		}

	} # for -- i

	result = sprintf("null")
	return result
}
#----------------------    Sc_EvalVar  -----------------------------------




function Sc_EvalRow(field,    i,j,k,ind,as,ary,mary,mrc,maxrc,maxrows,el,result) {
#
#	Evaluate a list of recursive ptrs as a row: build a matrix from the
#	top row downwards.
#
#	Since you can't return an array, the matrix returned will be
#	of the form, "a1 a2 a3... ; b1 b2 b3 ... ; ..."
#	----> "string-matrix" format
#

	#print "Sc_EvalRow receives (" field ")"

	#
	#	find the variable
	#
	for (i=1; i<=varcount; i++) {

		if (var_name[i] == field) {
			ind = i
		} # if -- field

	} # for -- i

	#

	#
	# parse the list
	#
	as = split(var_value[ind], ary, " ")
	if (as != var_size[ind]) {
		print "ERROR: Sc_EvalRow: num fields in list not equal size" | "cat 1>&2" 
		exit 1
	}

	#
	#	if it is a comment line, just ignore it
	#
	if (ary[1] == "c") {
		#print "Sc_EvalRow: returning c line: (" var_value[ind] ")"
		return var_value[ind]
	}

	#
	#	loop through list
	#
	for (i=1; i<=as; i++) {
		#
		#	if it is a number or rel-symb, cat it to the
		#	string "matrix", else pass it to Sc_EvalVar
		#

		if (is_a_numeric_var(ary[i])) {
			mary[i] = ary[i]
			#print "Sc_EvalRow: numeric mary " mary[i]
		}
		else if (is_a_relation_symbol(ary[i])) {
			mary[i] = ary[i]
			#print "Sc_EvalRow: relation mary " mary[i]
		}
		else {
			#
			#	the variable in ary[i] will be expanded
			#	into a string-matrix of literals
			#
			#	for now save this string-matrix in this
			#	array
			#
			mary[i] = Sc_EvalVar(ary[i])
			#print "Sc_EvalRow: from Sc_EvalVar:"  mary[i]

		} # else -- a variable

	} # for i in ary


	#
	#	get max dimensions
	#
	maxrows = 1
	for (i=1; i<=as; i++) {

		mrc = get_string_matrix_rowscols(mary[i])
		split(mrc, maxrc, " ")
		if (maxrc[1] > maxrows) {
			maxrows = maxrc[1]
		}

	}

	#
	#	loop down the rows building the final string-matrix
	#
	for (i=1; i<=maxrows; i++) {

		for (j=1; j<=as; j++) {

			mrc = get_string_matrix_rowscols(mary[j])
			split(mrc, maxrc, " ")
			#
			#	loop thru this submatrix columns
			#
			for (k=1; k<=maxrc[2]; k++) {
				#
				#	get the entry from this row,col
				#	if this row is beyond the max row of
				#	this submatrix, then repeat its last
				#	row entry
				#
				if (i <= maxrc[1]) {
					el = get_element_from_string_matrix(mary[j], i, k)
				}
				else {
					el = get_element_from_string_matrix(mary[j], maxrc[1], k)
				}
				#
				#	cat this element into this row of string
				#
				result = sprintf("%s %s", result, el)
			}

		} # for -- j in mary

		#
		#	next row--cat a ";" into the string
		#
		if (i < maxrows) {
			result = sprintf("%s ;", result)
		}

	} # for -- i


	return result
}
#------------------------  Sc_EvalRow  ----------------------------------




function Sc_EvalCol(field,    i,j,k,ind,as,ary,mary,mrc,maxrc,maxcols,el,result,ret) {
#
#	Evaluate a list of recursive ptrs as a column: build a matrix from the
#	left column towards the right
#
#	Since you can't return an array, the matrix returned will be
#	of the form, "a1 a2 a3... ; b1 b2 b3 ... ; ..."
#	----> "string-matrix" format
#

	#print "Sc_EvalCol receives (" field ")"

	#
	#	find the variable
	#
	for (i=1; i<=varcount; i++) {

		if (var_name[i] == field) {
			ind = i
		} # if -- field

	} # for -- i

	#

	#
	# parse the list
	#
	as = split(var_value[ind], ary, " ")
	if (as != var_size[ind]) {
		print "ERROR: Sc_EvalCol: num fields in list (" var_value[ind] ")  not equal size " var_size[ind] | "cat 1>&2" 
		exit 1
	}


	#
	#	if it is a comment line, just ignore it
	#
	if (ary[1] == "c") {
		#print "Sc_EvalCol: returning c line: (" var_value[ind] ")"
		return var_value[ind]
	}

	#
	#	loop through list
	#
	for (i=1; i<=as; i++) {
		#
		#	if it is a number or rel-symb, cat it to the
		#	string "matrix", else pass it to Sc_EvalVar
		#

		if (is_a_numeric_var(ary[i])) {
			mary[i] = ary[i]
			#print "numeric mary " mary[i]
		}
		else if (is_a_relation_symbol(ary[i])) {
			mary[i] = ary[i]
			#print "relation mary " mary[i]
		}
		else {
			#
			#	the variable in ary[i] will be expanded
			#	into a string-matrix of literals
			#
			#	for now save this string-matrix in this
			#	array
			#
			mary[i] = Sc_EvalVar(ary[i])
			#print "in Sc_EvalCol: Sc_EvalVar returns " mary[i]

		} # else -- a variable

	} # for i in ary


	#
	#	get max dimensions
	#
	maxcols = 1
	for (i=1; i<=as; i++) {

		mrc = get_string_matrix_rowscols(mary[i])
		split(mrc, maxrc, " ")
		if (maxrc[2] > maxcols) {
			maxcols = maxrc[2]
		}

	}

	#
	#	loop across the columns building the final string-matrix
	#
	for (i=1; i<=maxcols; i++) {

		for (j=1; j<=as; j++) {

			mrc = get_string_matrix_rowscols(mary[j])
			split(mrc, maxrc, " ")
			#
			#	loop thru this submatrix rows
			#
			for (k=1; k<=maxrc[1]; k++) {
				#
				#	get the entry from this row,col
				#	if this col is beyond the max col of
				#	this submatrix, then repeat its last
				#	col entry
				#
				if (i <= maxrc[2]) {
					el = get_element_from_string_matrix(mary[j], k, i)
				}
				else {
					el = get_element_from_string_matrix(mary[j], k, maxrc[1])
				}
				#
				#	cat this element into this col of string
				#	(this is done just like Sc_EvalRow, above
				#	because we will simply transpose the matrix before
				#	returning it)
				#
				result = sprintf("%s %s", result, el)
			}

		} # for -- j in mary

		#
		#	next row--cat a ";" into the string
		#
		if (i < maxcols) {
			result = sprintf("%s ;", result)
		}

	} # for -- i

	#
	#	now transpose the matrix
	#
	ret = transpose_string_matrix(result)

	return ret
}
#------------------------  Sc_EvalCol  ----------------------------------



function string_matrix_to_csound(sm,   i,j,as,ary,cas,cary,cmt,ltempo,isltempo,instr,stt,ltime,dur,outline)
{
#
#	after Sc_EvalVar crunches the variable on the score line
#	we convert it into csound score lines
#

	ltime = scoretime
	print "global score time is " scoretime

	#
	#	split into rows
	#
	as = split(sm, ary, ";")
	for (i=1; i<=as; i++) {

		#
		#	split this row into columns
		#
		cas = split(ary[i], cary, " ")

		#---------------------------------------------
		#	column 1
		#---------------------------------------------
		
		#
		#	the first column is always a tempo statement
		#	and doesn't become a column in the csound score
		#
		if (cary[1] == "c") {
			#
			#	a comment
			#	print everything after the "c" to the csound
			#	score
			#

			cmt = sprintf("; ")
			for (j=2; j<=cas; j++) {

				if (cary[j] == "c") {
					break
				}

				cmt = sprintf("%s %s", cmt, cary[j])

			} # for j
			print cmt >> "hold.sco"
			continue
		}

		#
		#	evaluate the first column as a tempo statement
		#
		if (is_a_numeric_var(cary[1])) {
			ltempo = cary[1]
			isltempo=1
		}
		else if (cary[1] == "&") {
			isltempo=0
		}
		else if (is_a_relation_symbol(cary[1])) {
			ltempo = get_relation(cary[1], tempo)
			isltempo=1
		}
		else {
			print "ERROR: string_matrix_to_csound: 1st column of a"\
				" variable expansion is not a tempo" | "cat 1>&2" 
			print "---> (" cary[1] ")" | "cat 1>&2" 
			print "in line (" ary[i] ")" | "cat 1>&2" 
			exit 1
		}

		#
		#	bug fix
		#	removed   "outline = ltempo"  9/27/2005
		#
		outline = ""

		#---------------------------------------------
		#	column 2	--  must be an instr index
		#---------------------------------------------
		#
		#	evaluate the second column as instrument statement
		#
		if (is_a_numeric_var(cary[2])) {
			instr = cary[2]
			
		}
		else {
			print "ERROR: string_matrix_to_csound: 2nd column of a"\
				" variable expansion is not legal for an instrument"\
				" index" | "cat 1>&2" 
			print "---> (" cary[2] ")" | "cat 1>&2" 
			print "in line (" ary[i] ")" | "cat 1>&2" 
			exit 1
		}

		outline = sprintf("%s i%g", outline, instr)


		#---------------------------------------------
		#	column 3	--  must be a start time
		#---------------------------------------------
		#
		#	evaluate the third column as a start time
		#
		if (is_a_numeric_var(cary[3])) {
			stt = cary[3]
		}
		else if (is_a_relation_symbol(cary[3])) {
			if (is_a_relation_scoretime(cary[3])) {
				stt = get_relation(cary[3], ltime)
			}
			else if (is_a_relation_ascoretime(cary[3])) {
				stt = get_relation(cary[3], scoretime)
			}
		}
		else {
			print "ERROR: string_matrix_to_csound: 3rd column of a"\
				" variable expansion is not legal for a start time" | "cat 1>&2" 
			print "---> (" cary[3] ")" | "cat 1>&2" 
			print "in line (" ary[i] ")" | "cat 1>&2" 
			exit 1
		}

		#
		#	(July 5, 2005)
		#	Before we print the start time to the output score line
		#	introduce a random offset if desired
		#
		if (use_random) {
			#
			#	random_range should be a fractional value between 0 and 1
			#	which is the percentage (+ or -) of a quater note, qdur
			#
			stt = stt + (qdur * random_range * 2 * (rand()-0.5))
		}

		#
		#	output the start time
		#
		outline = sprintf("%s %g", outline, stt)


		#---------------------------------------------
		#	column 4	--  must be a duration
		#---------------------------------------------
		#
		#	evaluate the fourth column as a duration
		#
		if (is_a_numeric_var(cary[4])) {
			dur = cary[4]
		}
		else if (is_a_relation_symbol(cary[4])) {
			if (isltempo) {
				#
				#	this line has a non-default tempo
				#
				dur = get_duration_with_ltempo(cary[4], ltempo)
			}
			else {
				dur = get_duration(cary[4])
			}
		}
		else {
			print "ERROR: string_matrix_to_csound: 4th column of a"\
				" variable expansion is not legal for a duration" | "cat 1>&2" 
			print "---> (" cary[4] ")" | "cat 1>&2" 
			print "in line (" ary[i] ")" | "cat 1>&2" 
			exit 1
		}

		outline = sprintf("%s %g", outline, dur)



		#
		#	now start reading from arbitrary assignments
		#	all variables should have been reduced to a number
		#	after the call to Sc_EvalVar
		#
		for (j=5; j<=cas; j++) {

			if (is_a_numeric_var(cary[j])) {
				outline = sprintf("%s %g", outline, cary[j])
			}

			else {
				print "ERROR: string_matrix_to_csound: column " j " of a"\
					" variable expansion is not legal" | "cat 1>&2" 
				print "---> (" cary[j] ")" | "cat 1>&2" 
				print "in line (" ary[i] ")" | "cat 1>&2" 
				exit 1
			}


		} # for -- cols

		#
		#
		#	print the line out to "hold.sco"
		#
		print outline >> "hold.sco"

		ltime += dur

	} # for -- rows

}
#--------------------- string_matrix_to_csound -------------------------




